from autosar.base import splitRef
from autosar.element import Element
import sys

class SystemSignal(Element):
    def __init__(self,name,dataTypeRef,initValueRef,length,desc=None,parent=None):
        super().__init__(name,parent)
        self.dataTypeRef=dataTypeRef
        self.initValueRef=initValueRef
        self.length=length
        self.desc=desc
        self.parent=parent

    def asdict(self):
        data={'type': self.__class__.__name__,'name':self.name,
              'dataTypeRef': self.dataTypeRef,
              'initValueRef': self.initValueRef,
              'length': self.length
              }
        if self.desc is not None: data['desc']=self.desc
        return data

class SystemSignalGroup(Element):
    def __init__(self, name, systemSignalRefs=None,parent=None):
        super().__init__(name,parent)
        if isinstance(systemSignalRefs,list):
            self.systemSignalRefs=systemSignalRefs
        else:
            self.systemSignalRefs=[]



import autosar.package
import autosar.element
import autosar.parser.parser_base
import sys
from autosar.base import parseTextNode

class PackageParser:
    def __init__(self,version):
        assert(isinstance(version, float))
        self.version=version
        self.registeredParsers={}
        self.switcher={}

    def registerElementParser(self, elementParser):
        """
        Registers a new element parser into the package parser
        """
        assert(isinstance(elementParser, autosar.parser.parser_base.ElementParser))
        name = type(elementParser).__name__
        if name not in self.registeredParsers:
            for tagname in elementParser.getSupportedTags():
                self.switcher[tagname]=elementParser
            self.registeredParsers[name] = elementParser

    def loadXML(self, package, xmlRoot):
        """
        Loads an XML package by repeatedly invoking its registered element parsers
        """
        assert(self.switcher is not None)
        if xmlRoot.find('ELEMENTS'):
            elementNames = set([x.name for x in package.elements])
            for xmlElement in xmlRoot.findall('./ELEMENTS/*'):
                parserObject = self.switcher.get(xmlElement.tag)
                if parserObject is not None:
                    element = parserObject.parseElement(xmlElement,package)
                    if element is None:
                        print("[PackageParser] No return value: %s"%xmlElement.tag)
                        continue
                    element.parent=package
                    if isinstance(element,autosar.element.Element):
                        if element.name not in elementNames:
                            #ignore duplicated items
                            package.append(element)
                            elementNames.add(element.name)
                    else:
                        #raise ValueError("parse error: %s"%type(element))
                        raise ValueError("parse error: %s"%xmlElement.tag)
                else:
                    package.unhandledParser.add(xmlElement.tag)

        if self.version >= 3.0 and self.version < 4.0:
            if xmlRoot.find('SUB-PACKAGES'):
                for xmlPackage in xmlRoot.findall('./SUB-PACKAGES/AR-PACKAGE'):
                    name = xmlPackage.find("./SHORT-NAME").text
                    subPackage = autosar.package.Package(name)
                    package.append(subPackage)
                    self.loadXML(subPackage,xmlPackage)
        elif self.version >= 4.0:
            for subPackageXML in xmlRoot.findall('./AR-PACKAGES/AR-PACKAGE'):
                name = parseTextNode(subPackageXML.find("./SHORT-NAME"))
                subPackage = package.find(name)
                if subPackage is None:
                    subPackage = autosar.package.Package(name)
                    package.append(subPackage)
                self.loadXML(subPackage, subPackageXML)



from autosar.writer.writer_base import ElementWriter
import autosar.base
import autosar.workspace
import autosar.constant
import autosar.mode
from autosar.writer.behavior_writer import XMLBehaviorWriter

class XMLComponentTypeWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)
        if version >= 4.0:
            self.behavior_writer=XMLBehaviorWriter(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                              'ApplicationSoftwareComponent': self.writeApplicationSoftwareComponentXML,
                              'ComplexDeviceDriverComponent': self.writeComplexDeviceDriverComponentXML,
                              'ServiceComponent': self.writeServiceComponentXML,
                              'SwcImplementation': self.writeSwcImplementationXML,
                              'CompositionComponent': self.writeCompositionComponentXML
            }
        elif self.version >= 4.0:
            self.switcher = {
                              'ApplicationSoftwareComponent': self.writeApplicationSoftwareComponentXML,
                              'ComplexDeviceDriverComponent': self.writeComplexDeviceDriverComponentXML,
                              'ServiceComponent': self.writeServiceComponentXML,
                              'SwcImplementation': self.writeSwcImplementationXML,
                              'CompositionComponent': self.writeCompositionComponentXML,
                              'ParameterComponent': self.writeParameterComponentXML,
                              'NvBlockComponent': self.writeNvBlockComponentXML
            }
        else:
            switch.keys = {}

    def getSupportedXML(self):
        return self.switcher.keys()

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        xmlWriteFunc = self.switcher.get(type(elem).__name__)
        if xmlWriteFunc is not None:
            return xmlWriteFunc(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeApplicationSoftwareComponentXML(self,swc):
        assert(isinstance(swc,autosar.component.ApplicationSoftwareComponent))
        return self._writeComponentXML(swc)

    def writeComplexDeviceDriverComponentXML(self,swc):
        assert(isinstance(swc,autosar.component.ComplexDeviceDriverComponent))
        return self._writeComponentXML(swc)

    def writeCompositionComponentXML(self,swc):
        assert(isinstance(swc,autosar.component.CompositionComponent))
        return self._writeComponentXML(swc)

    def writeServiceComponentXML(self,swc):
        assert(isinstance(swc,autosar.component.ServiceComponent))
        return self._writeComponentXML(swc)

    def writeParameterComponentXML(self, swc):
        assert(isinstance(swc,autosar.component.ParameterComponent))
        return self._writeComponentXML(swc)

    def writeNvBlockComponentXML(self, swc):
        assert(isinstance(swc,autosar.component.NvBlockComponent))
        return self._writeComponentXML(swc)

    def _writeComponentXML(self, swc):
        lines=[]
        ws = swc.rootWS()
        assert(ws is not None)
        lines=[]
        lines.append('<%s>'%swc.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%swc.name,1))
        if isinstance(swc, autosar.component.ServiceComponent):
            lines.append(self.indent('<CATEGORY>ServiceComponent</CATEGORY>',1))
        if swc.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(swc.adminData),1))
        descLines = self.writeDescXML(swc)
        if descLines is not None:
            lines.extend(self.indent(descLines,1))
        lines.append(self.indent('<PORTS>',1))
        for port in swc.providePorts:
            lines.extend(self.indent(self._writeProvidePortXML(port),2))
        for port in swc.requirePorts:
            lines.extend(self.indent(self._writeRequirePortXML(port),2))
        lines.append(self.indent('</PORTS>',1))
        if (self.version >= 4.0) and (isinstance(swc, autosar.component.AtomicSoftwareComponent)) and (swc.behavior is not None):
            lines.append(self.indent('<INTERNAL-BEHAVIORS>',1))
            lines.extend(self.indent(self.behavior_writer.writeInternalBehaviorXML(swc.behavior),2))
            lines.append(self.indent('</INTERNAL-BEHAVIORS>',1))
        if isinstance(swc, autosar.component.CompositionComponent):
            lines.extend(self.indent(self._writeComponentsXML(ws, swc.components),1))
            if (len(swc.assemblyConnectors)>0) or (len(swc.delegationConnectors)>0):
                lines.append(self.indent('<CONNECTORS>',1))
                if len(swc.assemblyConnectors)>0:
                    lines.extend(self.indent(self._writeAssemblyConnectorsXML(ws, swc.assemblyConnectors),2))
                if len(swc.delegationConnectors)>0:
                    lines.extend(self.indent(self._writeDelegationConnectorsXML(ws, swc.delegationConnectors),2))
                lines.append(self.indent('</CONNECTORS>',1))
        if isinstance(swc, autosar.component.NvBlockComponent):
            if (len(swc.nvBlockDescriptors) > 0):
                lines.append(self.indent('<NV-BLOCK-DESCRIPTORS>',1))
                for desc in swc.nvBlockDescriptors:
                    lines.extend(self.indent(self.behavior_writer.writeNvBlockDescriptorXML(desc),2))
                lines.append(self.indent('</NV-BLOCK-DESCRIPTORS>',1))
        lines.append('</%s>'%swc.tag(self.version))
        return lines


    def _writeRequirePortXML(self, port):
        lines=[]
        assert(port.ref is not None)
        ws=port.rootWS()
        assert(ws is not None)
        portInterface=ws.find(port.portInterfaceRef)
        if portInterface is None:
            raise autosar.base.InvalidPortInterfaceRef("{0.ref}: {0.portInterfaceRef}".format(port))
        lines.append('<R-PORT-PROTOTYPE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%port.name,1))
        if port.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(port.adminData),1))
        if isinstance(portInterface, autosar.portinterface.ClientServerInterface) and isinstance(port.parent, autosar.component.CompositionComponent) or len(port.comspec)==0:
            if self.version<4.0:
                lines.append(self.indent('<REQUIRED-COM-SPECS></REQUIRED-COM-SPECS>',1))

        else:
            lines.append(self.indent('<REQUIRED-COM-SPECS>',1))
            for comspec in port.comspec:
                if self.version>=4.0:
                    if isinstance(portInterface, autosar.portinterface.ModeSwitchInterface):
                        modeGroup = None
                        if comspec.name is not None:
                            modeGroup = portInterface.find(comspec.name)
                            if modeGroup is None:
                                raise ValueError("%s: Invalid comspec name '%s'"%(port.ref, comspec.name))
                        elif comspec.modeGroupRef is not None:
                            modeGroup = ws.find(comspec.modeGroupRef)
                            if modeGroup is None:
                                raise autosar.base.InvalidModeGroupRef(comspec.modeGroupRef)
                        lines.extend(self.indent(self._writeModeSwitchReceiverComSpecXML(ws, portInterface, comspec, modeGroup),2))
                    elif isinstance(portInterface, autosar.portinterface.ParameterInterface):
                        lines.extend(self.indent(self._writeParameterRequireComSpecXML(port, portInterface, comspec),2))
                    elif isinstance(portInterface, autosar.portinterface.SenderReceiverInterface):
                        dataElem=portInterface.find(comspec.name)
                        if dataElem is None:
                            raise ValueError("%s: Invalid comspec name '%s'"%(port.ref, comspec.name))
                        lines.extend(self.indent(self._writeDataReceiverComSpecXML(ws, dataElem, comspec),2))
                    elif isinstance(portInterface, autosar.portinterface.ClientServerInterface):
                        operation=portInterface.find(comspec.name)
                        if operation is None:
                            raise ValueError("%s: Invalid comspec name '%s'"%(port.ref,comspec.name))
                        lines.extend(self.indent(self._writeOperationComSpec(operation),2))
                    elif isinstance(portInterface, autosar.portinterface.NvDataInterface):
                        lines.extend(self.indent(self._writeNvDataRequireComSpecXML(port, ws, portInterface, comspec),2))
                    else:
                        raise NotImplementedError(str(type(portInterface)))
                else:
                    if isinstance(portInterface, autosar.portinterface.ClientServerInterface):
                        operation=portInterface.find(comspec.name)
                        if operation is None:
                            raise ValueError("%s: invalid comspec name '%s'"%(port.ref,comspec.name))
                        lines.extend(self.indent(self._writeOperationComSpec(operation),2))
                    elif isinstance(portInterface, autosar.portinterface.ParameterInterface):
                        pass #not supported in AUTOSAR 3
                    else:
                        dataElem=portInterface.find(comspec.name)
                        if dataElem is None:
                            raise ValueError("%s: invalid comspec name '%s'"%(port.ref, comspec.name))
                        lines.extend(self.indent(self._writeDataReceiverComSpecXML(ws, dataElem, comspec),2))
            lines.append(self.indent('</REQUIRED-COM-SPECS>',1))
        lines.append(self.indent('<REQUIRED-INTERFACE-TREF DEST="%s">%s</REQUIRED-INTERFACE-TREF>'%(portInterface.tag(self.version),portInterface.ref),1))
        lines.append('</R-PORT-PROTOTYPE>')
        return lines

    def _writeOperationComSpec(self, operation):
        lines = []
        lines.append('<CLIENT-COM-SPEC>')
        lines.append(self.indent('<OPERATION-REF DEST="%s">%s</OPERATION-REF>'%(operation.tag(self.version),operation.ref),1))
        lines.append('</CLIENT-COM-SPEC>')
        return lines

    def _writeDataReceiverComSpecXML(self, ws, dataElem, comspec):
        assert(isinstance(dataElem, autosar.portinterface.DataElement))
        lines = []
        if self.version<4.0:
            if dataElem.isQueued:
                lines.append('<QUEUED-RECEIVER-COM-SPEC>',)
                lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(dataElem.tag(self.version),dataElem.ref),1))
                lines.append(self.indent('<QUEUE-LENGTH>%d</QUEUE-LENGTH>'%(int(comspec.queueLength)),1))
                lines.append('</QUEUED-RECEIVER-COM-SPEC>')
            else:
                lines.append('<UNQUEUED-RECEIVER-COM-SPEC>')
                lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(dataElem.tag(self.version),dataElem.ref),1))
                lines.append(self.indent('<ALIVE-TIMEOUT>%d</ALIVE-TIMEOUT>'%(comspec.aliveTimeout),1))
                if comspec.initValueRef is not None:
                    tag = ws.find(comspec.initValueRef).tag(self.version)
                    lines.append(self.indent('<INIT-VALUE-REF DEST="%s">%s</INIT-VALUE-REF>'%(tag,comspec.initValueRef),1))
                lines.append('</UNQUEUED-RECEIVER-COM-SPEC>')
        else:
            if dataElem.isQueued:
                lines.append('<QUEUED-RECEIVER-COM-SPEC>')
                lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(dataElem.tag(self.version),dataElem.ref),1))
                lines.append(self.indent('<QUEUE-LENGTH>%d</QUEUE-LENGTH>'%(int(comspec.queueLength)),1))
                lines.append('</QUEUED-RECEIVER-COM-SPEC>')
            else:
                lines.append('<NONQUEUED-RECEIVER-COM-SPEC>')
                lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(dataElem.tag(self.version),dataElem.ref),1))
                lines.append(self.indent('<USES-END-TO-END-PROTECTION>false</USES-END-TO-END-PROTECTION>',1))

                lines.append(self.indent('<ALIVE-TIMEOUT>%d</ALIVE-TIMEOUT>'%(comspec.aliveTimeout),1))
                lines.append(self.indent('<ENABLE-UPDATE>false</ENABLE-UPDATE>',1))
                lines.append(self.indent('<FILTER>',1))
                lines.append(self.indent('<DATA-FILTER-TYPE>ALWAYS</DATA-FILTER-TYPE>',2))
                lines.append(self.indent('</FILTER>',1))
                lines.append(self.indent('<HANDLE-NEVER-RECEIVED>false</HANDLE-NEVER-RECEIVED>',1))
                if comspec.initValueRef is not None:
                    lines.extend(self.indent(self._writeInitValueRefXML(ws, comspec.initValueRef),1))
                if comspec.initValue is not None:
                    lines.append(self.indent('<INIT-VALUE>',1))
                    lines.extend(self.indent(self.writeValueSpecificationXML(comspec.initValue),2))
                    lines.append(self.indent('</INIT-VALUE>',1))
                lines.append('</NONQUEUED-RECEIVER-COM-SPEC>')
        return lines

    def _writeModeSwitchReceiverComSpecXML(self, ws, portInterface, comspec, modeGroup):
        lines = []
        lines.append('<MODE-SWITCH-RECEIVER-COM-SPEC>')
        if comspec.enhancedMode is not None:
            lines.append(self.indent('<ENHANCED-MODE-API>{}</ENHANCED-MODE-API>'.format(self.toBooleanStr(comspec.enhancedMode)),1))
        if modeGroup is not None:
            destTag = modeGroup.tag(self.version)
            if (self.version <= 4.2) and destTag == 'MODE-GROUP':
                #There is a bug in the XML Schema that needs a manual correction of its DEST value
                destTag = 'MODE-DECLARATION-GROUP-PROTOTYPE'
            lines.append(self.indent('<MODE-GROUP-REF DEST="{}">{}</MODE-GROUP-REF>'.format(destTag, modeGroup.ref), 1))
        if comspec.supportAsync is not None:
            lines.append(self.indent('<SUPPORTS-ASYNCHRONOUS-MODE-SWITCH>{}</SUPPORTS-ASYNCHRONOUS-MODE-SWITCH>'.format(self.toBooleanStr(comspec.supportAsync)),1))
        lines.append('</MODE-SWITCH-RECEIVER-COM-SPEC>')
        return lines

    def _writeParameterRequireComSpecXML(self, port, portInterface, comspec):
        lines = []
        parameter = portInterface.find(comspec.name)
        if parameter is None:
            raise ValueError('%s: invalid parameter reference name: %s'%(port.ref, comspec.name))
        lines.append('<PARAMETER-REQUIRE-COM-SPEC>')
        if comspec.initValue is not None:
            lines.append(self.indent('<INIT-VALUE>',1))
            lines.extend(self.indent(self.writeValueSpecificationXML(comspec.initValue),2))
            lines.append(self.indent('</INIT-VALUE>',1))
        lines.append(self.indent('<PARAMETER-REF DEST="%s">%s</PARAMETER-REF>'%(parameter.tag(self.version), parameter.ref),1))
        lines.append('</PARAMETER-REQUIRE-COM-SPEC>')
        return lines

    def _writeNvDataRequireComSpecXML(self, port, ws, portInterface, comspec):
        lines = []
        if self.version<4.0:
            raise NotImplementedError('_writeNvDataRequireComSpecXML')
        else:
            nvData = portInterface.find(comspec.name)
            if nvData is None:
                raise ValueError('%s: invalid nvData reference name: %s'%(port.ref, comspec.name))
            lines.append('<NV-REQUIRE-COM-SPEC>')
            if comspec.initValueRef is not None:
                lines.extend(self.indent(self._writeInitValueRefXML(ws, comspec.initValueRef),1))
            if comspec.initValue is not None:
                lines.append(self.indent('<INIT-VALUE>',1))
                lines.extend(self.indent(self.writeValueSpecificationXML(comspec.initValue),2))
                lines.append(self.indent('</INIT-VALUE>',1))
            lines.append(self.indent('<VARIABLE-REF DEST="%s">%s</VARIABLE-REF>'%(nvData.tag(self.version), nvData.ref),1))
            lines.append('</NV-REQUIRE-COM-SPEC>')
        return lines

    def _writeProvidePortXML(self, port):
        lines=[]
        assert(port.ref is not None)
        ws=port.rootWS()
        assert(ws is not None)
        portInterface=ws.find(port.portInterfaceRef)
        if portInterface is None:
            raise autosar.base.InvalidPortInterfaceRef("{0.ref}: {0.portInterfaceRef}".format(port))
        lines.append('<%s>'%port.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%port.name,1))
        if port.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(port.adminData),1))
        if isinstance(portInterface, autosar.portinterface.ClientServerInterface) and isinstance(port.parent, autosar.component.CompositionComponent) or len(port.comspec)==0:
            lines.append(self.indent('<PROVIDED-COM-SPECS></PROVIDED-COM-SPECS>',1))
        else:
            lines.append(self.indent('<PROVIDED-COM-SPECS>',1))
            for comspec in port.comspec:
                if isinstance(comspec, autosar.port.DataElementComSpec):
                    elem=portInterface.find(comspec.name)
                    if elem is None:
                        raise ValueError("%s: Invalid data element name '%s'"%(port.ref,comspec.name))
                    if elem.isQueued:
                        lines.extend(self.indent(self._writeQueuedSenderComSpecXML(ws, comspec, elem), 2))
                    else:
                        lines.extend(self.indent(self._writeUnqueuedSenderComSpecXML(ws, comspec, elem),2))
                elif isinstance(comspec, autosar.port.OperationComSpec):
                    operation=portInterface.find(comspec.name)
                    if operation is None:
                        raise ValueError("%s: Invalid operation name '%s'"%(port.ref, comspec.name))
                    lines.extend(self.indent(self._writeServerComSpecXML(comspec, operation), 2))
                elif isinstance(comspec, autosar.port.ParameterComSpec):
                    param=portInterface.find(comspec.name)
                    if param is None:
                        raise ValueError("%s: Invalid parameter name '%s'"%(port.ref, comspec.name))
                    lines.extend(self.indent(self._writeParameterProvideComSpecXML(ws, comspec, param),2))
                elif isinstance(comspec, autosar.port.ModeSwitchComSpec):
                    modeGroup = None
                    if comspec.name is not None:
                        modeGroup = portInterface.find(comspec.name)
                        if modeGroup is None:
                            raise ValueError("%s: Invalid mode group name '%s'"%(port.ref, comspec.name))
                    elif comspec.modeGroupRef is not None:
                        modeGroup = ws.find(comspec.modeGroupRef)
                        if modeGroup is None:
                            raise autosar.base.InvalidModeGroupRef(comspec.modeGroupRef)
                    lines.extend(self.indent(self._writeModeSwitchSenderComSpecXML(ws, comspec, modeGroup),2))
                elif isinstance(comspec, autosar.port.NvProvideComSpec):
                    nvProvideComSpec=portInterface.find(comspec.name)
                    if nvProvideComSpec is None:
                        raise ValueError("%s: Invalid parameter name '%s'"%(port.ref, comspec.name))
                    lines.extend(self.indent(self._writeNvDataProvideComSpecXML(port, ws, portInterface, comspec),2))
                else:
                    raise NotImplementedError(str(type(comspec)))
            lines.append(self.indent('</PROVIDED-COM-SPECS>',1))
        lines.append(self.indent('<PROVIDED-INTERFACE-TREF DEST="%s">%s</PROVIDED-INTERFACE-TREF>'%(portInterface.tag(self.version),portInterface.ref),1))
        lines.append('</%s>'%port.tag(self.version))
        return lines

    def _writeQueuedSenderComSpecXML(self, ws, comspec, elem):
        assert(isinstance(comspec, autosar.port.DataElementComSpec))
        lines=[]
        lines.append('<QUEUED-SENDER-COM-SPEC>')
        lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(elem.tag(self.version),elem.ref),1))
        if (self.version>=4.0 and comspec.useEndToEndProtection is not None):
            lines.append(self.indent('<USES-END-TO-END-PROTECTION>{}</USES-END-TO-END-PROTECTION>'.format(self.toBooleanStr(comspec.useEndToEndProtection)), 1))
        lines.append('</QUEUED-SENDER-COM-SPEC>')
        return lines

    def _writeUnqueuedSenderComSpecXML(self, ws, comspec, elem):
        lines=[]
        if self.version<4.0:
            lines.append('<UNQUEUED-SENDER-COM-SPEC>')
            lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(elem.tag(self.version),elem.ref),1))
            if isinstance(comspec.canInvalidate,bool):
                lines.append(self.indent('<CAN-INVALIDATE>%s</CAN-INVALIDATE>'%('true' if comspec.canInvalidate else 'false'),1))
            if comspec.initValueRef is not None:
                tag = ws.find(comspec.initValueRef).tag(self.version)
                lines.append(self.indent('<INIT-VALUE-REF DEST="%s">%s</INIT-VALUE-REF>'%(tag,comspec.initValueRef),1))
            lines.append('</UNQUEUED-SENDER-COM-SPEC>')
        else:
            lines.append('<NONQUEUED-SENDER-COM-SPEC>')
            lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(elem.tag(self.version),elem.ref),1))
            lines.append(self.indent('<USES-END-TO-END-PROTECTION>false</USES-END-TO-END-PROTECTION>',1))
            if comspec.initValueRef is not None:
                lines.extend(self.indent(self._writeInitValueRefXML(ws, comspec.initValueRef),1))
            if comspec.initValue is not None:
                lines.append(self.indent('<INIT-VALUE>',1))
                lines.extend(self.indent(self.writeValueSpecificationXML(comspec.initValue),2))
                lines.append(self.indent('</INIT-VALUE>',1))
            lines.append('</NONQUEUED-SENDER-COM-SPEC>')
        return lines

    def _writeServerComSpecXML(self, comspec, operation):
        assert(isinstance(comspec, autosar.port.OperationComSpec))
        lines=[]
        lines.append('<SERVER-COM-SPEC>')
        lines.append(self.indent('<OPERATION-REF DEST="%s">%s</OPERATION-REF>'%(operation.tag(self.version),operation.ref),1))
        lines.append(self.indent('<QUEUE-LENGTH>%d</QUEUE-LENGTH>'%(int(comspec.queueLength)),1))
        lines.append('</SERVER-COM-SPEC>')
        return lines

    def _writeParameterProvideComSpecXML(self, ws, comspec, param):
        assert(isinstance(comspec, autosar.port.ParameterComSpec))
        lines = []
        if self.version<4.0:
            raise NotImplementedError('_writeParameterProvideComSpecXML')
        else:
            lines.append('<PARAMETER-PROVIDE-COM-SPEC>')
            if comspec.initValue is not None:
                lines.append(self.indent('<INIT-VALUE>',1))
                lines.extend(self.indent(self.writeValueSpecificationXML(comspec.initValue),2))
                lines.append(self.indent('</INIT-VALUE>',1))
            lines.append(self.indent('<PARAMETER-REF DEST="%s">%s</PARAMETER-REF>'%(param.tag(self.version),param.ref),1))
            lines.append('</PARAMETER-PROVIDE-COM-SPEC>')
        return lines

    def _writeNvDataProvideComSpecXML(self, port, ws, portInterface, comspec):
        lines = []
        if self.version<4.0:
            raise NotImplementedError('_writeNvDataProvideComSpecXML')
        else:
            nvData = portInterface.find(comspec.name)
            if nvData is None:
                raise ValueError('%s: invalid nvData reference name: %s'%(port.ref, comspec.name))
            lines.append('<NV-PROVIDE-COM-SPEC>')
            if comspec.ramBlockInitValueRef is not None:
                constant = ws.find(comspec.ramBlockInitValueRef)
                if constant is None:
                    raise ValueError('Invalid constant reference: %s'%comspec.ramBlockInitValueRef)
                lines.append(self.indent('<RAM-BLOCK-INIT-VALUE>',1))
                lines.append(self.indent('<CONSTANT-REFERENCE>',2))
                lines.append(self.indent('<CONSTANT-REF DEST="%s">%s</CONSTANT-REF>'%(constant.tag(self.version),constant.ref),3))
                lines.append(self.indent('</CONSTANT-REFERENCE>',2))
                lines.append(self.indent('</RAM-BLOCK-INIT-VALUE>',1))
            if comspec.ramBlockInitValue is not None:
                lines.append(self.indent('<RAM-BLOCK-INIT-VALUE>',1))
                lines.extend(self.indent(self.writeValueSpecificationXML(comspec.ramBlockInitValue),2))
                lines.append(self.indent('</RAM-BLOCK-INIT-VALUE>',1))
            if comspec.romBlockInitValueRef is not None:
                constant = ws.find(comspec.romBlockInitValueRef)
                if constant is None:
                    raise ValueError('Invalid constant reference: %s'%comspec.romBlockInitValueRef)
                lines.append(self.indent('<ROM-BLOCK-INIT-VALUE>',1))
                lines.append(self.indent('<CONSTANT-REFERENCE>',2))
                lines.append(self.indent('<CONSTANT-REF DEST="%s">%s</CONSTANT-REF>'%(constant.tag(self.version),constant.ref),3))
                lines.append(self.indent('</CONSTANT-REFERENCE>',2))
                lines.append(self.indent('</ROM-BLOCK-INIT-VALUE>',1))
            if comspec.romBlockInitValue is not None:
                lines.append(self.indent('<ROM-BLOCK-INIT-VALUE>',1))
                lines.extend(self.indent(self.writeValueSpecificationXML(comspec.romBlockInitValue),2))
                lines.append(self.indent('</ROM-BLOCK-INIT-VALUE>',1))
            lines.append(self.indent('<VARIABLE-REF DEST="%s">%s</VARIABLE-REF>'%(nvData.tag(self.version), nvData.ref),1))
            lines.append('</NV-PROVIDE-COM-SPEC>')
        return lines

    def _writeModeSwitchSenderComSpecXML(self, ws, comspec, modeGroup):
        assert(isinstance(comspec, autosar.port.ModeSwitchComSpec))
        lines=[]
        lines.append('<MODE-SWITCH-SENDER-COM-SPEC>')
        if comspec.enhancedMode is not None:
            lines.append(self.indent('<ENHANCED-MODE-API>{}</ENHANCED-MODE-API>'.format(self.toBooleanStr(comspec.enhancedMode)),1))
        if modeGroup is not None:
            destTag = modeGroup.tag(self.version)
            if (self.version <= 4.2) and destTag == 'MODE-GROUP':
                #There is a bug in the XML Schema that needs a custom DEST value
                destTag = 'MODE-DECLARATION-GROUP-PROTOTYPE'
            lines.append(self.indent('<MODE-GROUP-REF DEST="{}">{}</MODE-GROUP-REF>'.format(destTag, modeGroup.ref), 1))
        if comspec.modeSwitchAckTimeout is not None:
            timeoutStr = self.format_float(float(comspec.modeSwitchAckTimeout)/1000.0)
            lines.append(self.indent('<MODE-SWITCHED-ACK>',1))
            lines.append(self.indent('<TIMEOUT>{}</TIMEOUT>'.format(timeoutStr),2))
            lines.append(self.indent('</MODE-SWITCHED-ACK>',1))
        if comspec.queueLength is not None:
            lines.append(self.indent('<QUEUE-LENGTH>{:d}</QUEUE-LENGTH>'.format(int(comspec.queueLength)),1))
        lines.append('</MODE-SWITCH-SENDER-COM-SPEC>')
        return lines

    def writeSwcImplementationXML(self, elem):
        assert(isinstance(elem,autosar.component.SwcImplementation))
        lines=[]
        ws = elem.rootWS()
        assert(ws is not None)

        lines=['<SWC-IMPLEMENTATION>']
        lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SHORT-NAME', text=elem.name), 1))
        if elem.codeDescriptors is not None:

            lines.append(self.indent('<CODE-DESCRIPTORS>',1))
            for decriptor in elem.codeDescriptors:

                lines.append(self.indent('<CODE>',2))
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SHORT-NAME', text=decriptor.name), 3))

                if self.version >= 4.0:
                    if decriptor.artifactDescriptors is not None:
                        lines.append(self.indent('<ARTIFACT-DESCRIPTORS>',3))
                        for artifact in decriptor.artifactDescriptors:

                            lines.append(self.indent('<AUTOSAR-ENGINEERING-OBJECT>',4))
                            if artifact.shortLabel is not None:
                                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SHORT-LABEL', text=artifact.shortLabel), 5))

                            if artifact.category is not None:
                                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='CATEGORY', text=artifact.category), 5))

                            if artifact.revisionLabels is not None:
                                lines.append(self.indent('<REVISION-LABELS>',5))
                                for revisionLabel in artifact.revisionLabels:
                                    lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='REVISION-LABEL', text=revisionLabel), 6))
                                lines.append(self.indent('</REVISION-LABELS>',5))

                            if artifact.domain is not None:
                                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='DOMAIN', text=artifact.domain), 5))

                            lines.append(self.indent('</AUTOSAR-ENGINEERING-OBJECT>',4))

                        lines.append(self.indent('</ARTIFACT-DESCRIPTORS>',3))
                else:
                    if decriptor.type is not None:
                        lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='TYPE', text=decriptor.type), 3))
                lines.append(self.indent('</CODE>',2))
            lines.append(self.indent('</CODE-DESCRIPTORS>',1))

        if self.version > 4.0:
            if elem.programmingLanguage is not None:
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='PROGRAMMING-LANGUAGE', text=elem.programmingLanguage), 1))

            if elem.resourceConsumption is not None:
                lines.append(self.indent('<RESOURCE-CONSUMPTION>', 1))
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SHORT-NAME', text=elem.resourceConsumption.name), 2))
                lines.extend(self.indent(self._writeMemorySectionXML(ws, elem.resourceConsumption.memorySections), 2))
                lines.append(self.indent('</RESOURCE-CONSUMPTION>', 1))

            if elem.swVersion is not None:
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SW-VERSION', text=elem.swVersion), 1))

            if elem.useCodeGenerator is not None:
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='USED-CODE-GENERATOR', text=elem.useCodeGenerator), 1))

            if elem.vendorId is not None:
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='VENDOR-ID', text=elem.vendorId), 1))

        # Find the behavior to reference.
        behavior = ws.find(elem.behaviorRef)
        if behavior is None:
            raise autosar.base.InvalidBehaviorRef(elem.behaviorRef)

        if self.version < 4.0:
            ref = elem.behaviorRef
        else:
            ref = behavior.ref
        lines.append(self.indent('<{tag} DEST="{DEST}">{text}</{tag}>'.format(tag='BEHAVIOR-REF', DEST=behavior.tag(self.version), text=ref), 1))

        lines.append('</SWC-IMPLEMENTATION>')
        return lines

    def _writeMemorySectionXML(self, ws, memorySections):
        lines=[]
        if memorySections is not None:
            lines.append('<MEMORY-SECTIONS>')
            for memorySection in memorySections:
                lines.append(self.indent('<MEMORY-SECTION>', 1))
                lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SHORT-NAME', text=memorySection.name), 2))

                if memorySection.aligment is not None:
                    lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='ALIGNMENT', text=memorySection.aligment), 2))

                if memorySection.memClassSymbol is not None:
                    lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='MEM-CLASS-SYMBOL', text=memorySection.memClassSymbol), 2))

                if memorySection.options is not None:
                    lines.append(self.indent('<OPTIONS>', 2))
                    for option in memorySection.options:
                        lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='OPTION', text=option), 3))
                    lines.append(self.indent('</OPTIONS>', 2))

                if memorySection.size is not None:
                    lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SIZE', text=memorySection.size), 2))

                if memorySection.swAddrmethodRef is not None:
                    swc = ws.find(memorySection.swAddrmethodRef)
                    if swc is None:
                        raise autosar.base.InvalidSwAddrmethodRef(memorySection.swAddrmethodRef)
                    lines.append(self.indent('<{tag} DEST="{dest}">{text}</{tag}>'.format(tag='SW-ADDRMETHOD-REF', dest=swc.tag(self.version), text=swc.ref), 2))

                if memorySection.symbol is not None:
                    lines.append(self.indent('<{tag}>{text}</{tag}>'.format(tag='SYMBOL', text=memorySection.symbol), 2))

                lines.append(self.indent('</MEMORY-SECTION>', 1))
            lines.append('</MEMORY-SECTIONS>')
        return lines

    def _writeComponentsXML(self, ws, components):
        lines=[]
        if len(components)>0:
            lines.append('<COMPONENTS>')
            componentTag = 'SW-COMPONENT-PROTOTYPE' if self.version >= 4.0 else 'COMPONENT-PROTOTYPE'
            for component in components:
                swc = ws.find(component.typeRef)
                if swc is None:
                    raise ValueError('Invalid reference: '+component.typeRef)
                lines.append(self.indent('<%s>'%componentTag,1))
                lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%component.name,2))
                lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(swc.tag(self.version),swc.ref),2))
                lines.append(self.indent('</%s>'%componentTag,1))
            lines.append('</COMPONENTS>')
        return lines

    def _writeAssemblyConnectorsXML(self, ws, connectors):
        lines=[]
        for connector in connectors:
            lines.append('<%s>'%connector.tag(self.version))
            lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%connector.name, 1))
            lines.append(self.indent('<%s>'%connector.providerInstanceRef.tag(self.version), 1))
            providerComponent = ws.find(connector.providerInstanceRef.componentRef)
            if providerComponent is None:
                raise ValueError('invalid reference: ' +connector.providerInstanceRef.componentRef)
            providePort = ws.find(connector.providerInstanceRef.portRef)
            if providePort is None:
                raise ValueError('invalid reference: ' +connector.providerInstanceRef.portRef)
            componentTag = 'CONTEXT-COMPONENT-REF' if self.version >= 4.0 else 'COMPONENT-PROTOTYPE-REF'
            portTag = 'TARGET-P-PORT-REF' if self.version >= 4.0 else 'P-PORT-PROTOTYPE-REF'
            lines.append(self.indent('<%s DEST="%s">%s</%s>'%(componentTag, providerComponent.tag(self.version), providerComponent.ref, componentTag), 2))
            lines.append(self.indent('<%s DEST="%s">%s</%s>'%(portTag, providePort.tag(self.version), providePort.ref, portTag), 2))
            lines.append(self.indent('</%s>'%connector.providerInstanceRef.tag(self.version), 1))
            lines.append(self.indent('<%s>'%connector.requesterInstanceRef.tag(self.version), 1))
            requesterComponent = ws.find(connector.requesterInstanceRef.componentRef)
            if requesterComponent is None:
                raise ValueError('invalid reference: ' +connector.requesterInstanceRef.componentRef)
            requirePort = ws.find(connector.requesterInstanceRef.portRef)
            if requirePort is None:
                raise ValueError('invalid reference: ' +connector.requesterInstanceRef.portRef)
            portTag = 'TARGET-R-PORT-REF' if self.version >= 4.0 else 'R-PORT-PROTOTYPE-REF'
            lines.append(self.indent('<%s DEST="%s">%s</%s>'%(componentTag, requesterComponent.tag(self.version), requesterComponent.ref, componentTag), 2))
            lines.append(self.indent('<%s DEST="%s">%s</%s>'%(portTag, requirePort.tag(self.version), requirePort.ref, portTag), 2))
            lines.append(self.indent('</%s>'%connector.requesterInstanceRef.tag(self.version), 1))
            lines.append('</%s>'%connector.tag(self.version))
        return lines

    def _writeDelegationConnectorsXML(self, ws, connectors):
        lines=[]
        for connector in connectors:
            lines.append('<%s>'%connector.tag(self.version))
            lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%connector.name, 1))
            lines.append(self.indent('<%s>'%connector.innerPortInstanceRef.tag(self.version), 1))
            innerComponent = ws.find(connector.innerPortInstanceRef.componentRef)
            if innerComponent is None:
                raise ValueError('invalid reference: ' +connector.innerPortInstanceRef.componentRef)
            innerPort = ws.find(connector.innerPortInstanceRef.portRef)
            if innerPort is None:
                raise ValueError('invalid reference: ' +connector.innerPortInstanceRef.portRef)
            if self.version >= 4.0:
                lines.extend(self.indent(self._writeInnerPortRefV4(innerComponent,innerPort),2))
            else:
                lines.extend(self.indent(self._writeInnerPortRefV3(innerComponent,innerPort),2))
            lines.append(self.indent('</%s>'%connector.innerPortInstanceRef.tag(self.version), 1))
            outerPort = ws.find(connector.outerPortRef.portRef)
            if outerPort is None:
                raise ValueError('invalid reference: ' +connector.outerPortRef.portRef)
            lines.append(self.indent('<OUTER-PORT-REF DEST="%s">%s</OUTER-PORT-REF>'%(outerPort.tag(self.version), outerPort.ref), 1))
            lines.append('</%s>'%connector.tag(self.version))
        return lines

    def _writeInnerPortRefV3(self, innerComponent, innerPort):
        lines = []
        lines.append('<COMPONENT-PROTOTYPE-REF DEST="%s">%s</COMPONENT-PROTOTYPE-REF>'%(innerComponent.tag(self.version), innerComponent.ref))
        lines.append('<PORT-PROTOTYPE-REF DEST="%s">%s</PORT-PROTOTYPE-REF>'%(innerPort.tag(self.version), innerPort.ref))
        return lines

    def _writeInnerPortRefV4(self, innerComponent, innerPort):
        lines = []
        if isinstance(innerPort, autosar.port.RequirePort):
            outerTag = 'R-PORT-IN-COMPOSITION-INSTANCE-REF'
            innerTag = 'TARGET-R-PORT-REF'
        else:
            outerTag = 'P-PORT-IN-COMPOSITION-INSTANCE-REF'
            innerTag = 'TARGET-P-PORT-REF'
        lines.append('<%s>'%outerTag)
        lines.append(self.indent('<CONTEXT-COMPONENT-REF DEST="%s">%s</CONTEXT-COMPONENT-REF>'%(innerComponent.tag(self.version), innerComponent.ref), 1))
        lines.append(self.indent('<%s DEST="%s">%s</%s>'%(innerTag, innerPort.tag(self.version), innerPort.ref, innerTag), 1))
        lines.append('</%s>'%outerTag)
        return lines

    def _writeInitValueRefXML(self, ws, initValueRef):
        lines = []
        constant = ws.find(initValueRef)
        if constant is None:
            raise ValueError('Invalid constant reference: %s'%initValueRef)
        lines.append('<INIT-VALUE>')
        lines.append(self.indent('<CONSTANT-REFERENCE>',1))
        lines.append(self.indent('<CONSTANT-REF DEST="%s">%s</CONSTANT-REF>'%(constant.tag(self.version),constant.ref),2))
        lines.append(self.indent('</CONSTANT-REFERENCE>',1))
        lines.append('</INIT-VALUE>')
        return lines

class CodeComponentTypeWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                             'ApplicationSoftwareComponent': self.writeApplicationSoftwareComponentCode,
                             'ComplexDeviceDriverComponent': self.writeComplexDeviceDriverComponentCode,
                             'ServiceComponent': self.writeServiceComponentCode,
                             'ParameterComponent': self.writeParameterComponentCode,
                             'SwcImplementation': self.writeSwcImplementationCode,
                             'CompositionComponent': self.writeCompositionComponentCode,
            }
        elif self.version >= 4.0:
            self.switcher = {
            }
        else:
            switch.keys = {}

    def getSupportedXML(self):
        return []

    def getSupportedCode(self):
        return self.switcher.keys()

    def writeElementXML(self, elem):
        raise NotImplementedError('writeElementXML')

    def writeElementCode(self, elem, localvars):
        codeWriteFunc = self.switcher.get(type(elem).__name__)
        if codeWriteFunc is not None:
            return codeWriteFunc(elem, localvars)
        else:
            return None

    def writeApplicationSoftwareComponentCode(self, swc, localvars):
        return self._writeComponentCode(swc, 'createApplicationSoftwareComponent', localvars)

    def writeComplexDeviceDriverComponentCode(self, swc, localvars):
        return self._writeComponentCode(swc, 'createComplexDeviceDriverComponent', localvars)

    def writeServiceComponentCode(self, swc, localvars):
        return self._writeComponentCode(swc, 'createServiceComponent', localvars)

    def writeCompositionComponentCode(self, swc, localvars):
        return self._writeComponentCode(swc, 'createCompositionComponent', localvars)

    def writeParameterComponentCode(self, swc, localvars):
        return self._writeComponentCode(swc, 'createParameterComponent', localvars)


    def _writeComponentCode(self, swc, methodName, localvars):
        lines=[]
        ws = localvars['ws']
        lines.append("swc = package.%s(%s)"%(methodName,repr(swc.name)))
        localvars['swc']=swc
        if len(swc.providePorts)>0:
            lines.extend(self._writeComponentProvidePortsCode(swc))
        if len(swc.requirePorts)>0:
            lines.extend(self._writeComponentRequirePortsCode(swc))
        if isinstance(swc, autosar.component.CompositionComponent):
            if len(swc.components)>0:
                lines.extend(self._writeComponentsCode(swc.components, localvars))
            if len(swc.assemblyConnectors)>0:
                lines.extend(self._writeAssemblyConnectorsCode(swc.assemblyConnectors, localvars))
            if len(swc.delegationConnectors)>0:
                lines.extend(self._writeDelegationConnectorsCode(swc.delegationConnectors, localvars))
        return lines

    def _writeComponentProvidePortsCode(self,swc):
        lines=[]
        ws=swc.rootWS()
        assert(isinstance(ws,autosar.Workspace))
        for port in swc.providePorts:
            params=[]
            #name
            params.append(repr(port.name))
            portInterface=ws.find(port.portInterfaceRef, role='PortInterface')
            if portInterface is None:
                raise ValueError('invalid reference: '+port.portInterfaceRef)
            if isinstance(portInterface,autosar.portinterface.PortInterface):
                #portInterfaceRef
                if ws.roles['PortInterface'] is not None:
                    params.append(repr(portInterface.name)) #use name only
                else:
                    params.append(repr(portInterface.ref)) #use full reference
                if isinstance(portInterface,autosar.portinterface.SenderReceiverInterface):
                    #comspec
                    if len(port.comspec)>0:
                        if len(port.comspec)==1:
                            comspec=port.comspec[0]
                            if comspec.initValueRef is not None:
                                initValue = ws.find(comspec.initValueRef)
                                if initValue is None:
                                    raise ValueError('invalid reference: '+comspec.initValueRef)
                                if isinstance(initValue, autosar.constant.Constant):
                                    pass
                                elif isinstance(initValue, autosar.constant.Value):
                                    initValue = initValue.parent #we can shorten the ref string by pointing to its direct parent instead
                                else:
                                    raise ValueError('invalid reference type "%s" for "%s", expected Constant or Value'%(str(type(initValue)), initValue.ref))
                                if ws.roles['Constant'] is not None:
                                    params.append('initValueRef='+repr(initValue.name)) #use name only
                                else:
                                    params.append('initValueRef='+repr(initValue.ref)) #use full reference
                        else:
                            raise NotImplementedError('multiple comspecs not yet supported')
            else:
                raise NotImplementedError(type(portInterface))
            lines.append('swc.createProvidePort(%s)'%(', '.join(params)))
        return lines

    def _writeComponentRequirePortsCode(self,swc):
        lines=[]
        ws=swc.rootWS()
        assert(isinstance(ws,autosar.Workspace))
        for port in swc.requirePorts:
            params=[]
            #name
            params.append(repr(port.name))
            portInterface=ws.find(port.portInterfaceRef, role='PortInterface')
            if portInterface is None:
                raise ValueError('invalid reference: '+port.portInterfaceRef)
            if isinstance(portInterface,autosar.portinterface.PortInterface):
                #portInterfaceRef
                if ws.roles['PortInterface'] is not None:
                    params.append(repr(portInterface.name)) #use name only
                else:
                    params.append(repr(portInterface.ref)) #use full reference
                if isinstance(portInterface,autosar.portinterface.SenderReceiverInterface):
                    if len(port.comspec)>0:
                        if len(port.comspec)==1:
                            comspec=port.comspec[0]
                            if comspec.initValueRef is not None:
                                initValue = ws.find(comspec.initValueRef)
                                if initValue is None:
                                    raise ValueError('invalid reference: '+comspec.initValueRef)
                                if isinstance(initValue, autosar.constant.Constant):
                                    pass
                                elif isinstance(initValue, autosar.constant.Value):
                                    initValue = initValue.parent #we can shorten the ref string by pointing to its direct parent instead
                                else:
                                    raise ValueError('invalid reference type "%s" for "%s", expected Constant or Value'%(str(type(initValue)), initValue.ref))
                                if ws.roles['Constant'] is not None:
                                    params.append('initValueRef='+repr(initValue.name)) #use name only
                                else:
                                    params.append('initValueRef='+repr(initValue.ref)) #use full reference
                            if (comspec.aliveTimeout is not None) and int(comspec.aliveTimeout)>0:
                                params.append("aliveTimeout=%d"%int(comspec.aliveTimeout))
                            if (comspec.queueLength is not None) and int(comspec.queueLength)!=1:
                                params.append("queueLength=%d"%int(comspec.queueLength))
                        else:
                            raise NotImplementedError('multiple comspecs not yet supported')
            else:
                raise NotImplementedError(type(portInterface))
            lines.append('swc.createRequirePort(%s)'%(', '.join(params)))
        return lines


    def writeSwcImplementationCode(self, elem, localvars):
        """
        No need to generate python code for explicitly creating SwcImplementation object
        """
        return []


    def _writeComponentsCode(self, components, localvars):
        lines=[]
        assert('swc' in localvars)
        for component in components:
            params=[]
            assert(isinstance(component, autosar.component.ComponentPrototype))
            params.append(repr(self._createComponentRef(component.typeRef, localvars)))
            lines.append('swc.createComponentRef(%s)'%(', '.join(params)))
        return lines

    def _writeAssemblyConnectorsCode(self, connectors, localvars):
        lines=[]
        ws = localvars['ws']
        swc = localvars['swc']

        assert(ws is not None)
        for connector in connectors:
            params=[]
            assert(isinstance(connector, autosar.component.AssemblyConnector))
            providePort = ws.find(connector.providerInstanceRef.portRef)
            if providePort is None:
                raise ValueError('invalid reference: ' + connector.providerInstanceRef.portRef)
            provideCompRef = self._createComponentRef(providePort.parent.ref, localvars)
            requirePort = ws.find(connector.requesterInstanceRef.portRef)
            if requirePort is None:
                raise ValueError('invalid reference: ' + connector.requesterInstanceRef.portRef)
            requireCompRef = self._createComponentRef(requirePort.parent.ref, localvars)

            params.append(repr('%s/%s'%(provideCompRef,providePort.name)))
            params.append(repr('%s/%s'%(requireCompRef,requirePort.name)))
            lines.append('swc.createConnector(%s)'%(', '.join(params)))
        return lines

    def _writeDelegationConnectorsCode(self, connectors, localvars):
        lines=[]
        ws = localvars['ws']
        swc = localvars['swc']

        assert(ws is not None)
        for connector in connectors:
            params=[]
            assert(isinstance(connector, autosar.component.DelegationConnector))
            innerPort = ws.find(connector.innerPortInstanceRef.portRef)
            if innerPort is None:
                raise ValueError('invalid reference: ' + connector.innerPortInstanceRef.portRef)
            innerComponentRef = self._createComponentRef(innerPort.parent.ref, localvars)

            parts = autosar.base.splitRef(connector.outerPortRef.portRef)
            outerPort = swc.find(parts[-1])
            if outerPort is None or not isinstance(outerPort, autosar.port.Port):
                raise ValueError('no port with name "%s" found in Component %s'%(parts[-1], swc.ref))
            params.append(repr(outerPort.name))
            params.append(repr('%s/%s'%(innerComponentRef,innerPort.name)))
            lines.append('swc.createConnector(%s)'%(', '.join(params)))
        return lines



import sys
from autosar.parser.parser_base import ElementParser
import autosar.datatype

class ModeDeclarationParser(ElementParser):
    def __init__(self,version=3):
        self.version=version

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {'MODE-DECLARATION-GROUP': self.parseModeDeclarationGroup,
                             'MODE-DECLARATIONS': self.parseModeDeclarations
            }
        elif self.version >= 4.0:
            self.switcher = {
               'MODE-DECLARATION-GROUP': self.parseModeDeclarationGroup
            }

    def getSupportedTags(self):
        return self.switcher.keys()

    def parseElement(self, xmlElement, parent = None):
        parseFunc = self.switcher.get(xmlElement.tag)
        if parseFunc is not None:
            return parseFunc(xmlElement,parent)
        else:
            return None

    def parseModeDeclarationGroup(self,xmlRoot,rootProject=None,parent=None):
        assert(xmlRoot.tag == 'MODE-DECLARATION-GROUP')
        name = self.parseTextNode(xmlRoot.find("./SHORT-NAME"))
        category = self.parseTextNode(xmlRoot.find("./CATEGORY"))
        initialModeRef = self.parseTextNode(xmlRoot.find('./INITIAL-MODE-REF'))
        modeDclrGroup = autosar.mode.ModeDeclarationGroup(name,initialModeRef,None,parent)
        if xmlRoot.find('./MODE-DECLARATIONS') is not None:
            self.parseModeDeclarations(xmlRoot.find('./MODE-DECLARATIONS'), modeDclrGroup)
        if self.hasAdminData(xmlRoot):
            adminData = self.parseAdminDataNode(xmlRoot.find('ADMIN-DATA'))
            modeDclrGroup.adminData = adminData
        if category is not None:
            modeDclrGroup.category = category
        return modeDclrGroup

    def parseModeDeclarations(self,xmlRoot,parent):
        assert(xmlRoot.tag=="MODE-DECLARATIONS")
        assert(isinstance(parent, autosar.mode.ModeDeclarationGroup))
        result = []
        for mode in xmlRoot.findall("./MODE-DECLARATION"):
            declarationName = self.parseTextNode(mode.find("./SHORT-NAME"))
            declarationValue = None
            declarationValueXML = mode.find("./VALUE")
            if declarationValueXML is not None:
                declarationValue = self.parseTextNode(declarationValueXML)
            parent.modeDeclarations.append(autosar.mode.ModeDeclaration(declarationName, declarationValue, parent = parent))
        return result




from autosar.base import parseXMLFile,splitRef,parseTextNode,parseIntNode,hasAdminData,parseAdminDataNode
from autosar.system import *
from autosar.parser.parser_base import ElementParser

class SystemParser(ElementParser):
    def __init__(self,version=3.0):
        super().__init__(version)

    def getSupportedTags(self):
        return ['SYSTEM']

    def parseElement(self, xmlElement, parent = None):
        if xmlElement.tag == 'SYSTEM':
            return self.parseSystem(xmlElement, parent)
        else:
            return None

    def parseSystem(self,xmlRoot,parent=None):
        """
        parses <SYSTEM>
        """
        assert(xmlRoot.tag=='SYSTEM')
        xmlName = xmlRoot.find('SHORT-NAME')
        if xmlName is not None:
            system=System(parseTextNode(xmlName),parent)
            for xmlElem in xmlRoot.findall('./*'):
                if xmlElem.tag=='SHORT-NAME':
                    pass
                elif xmlElem.tag=='ADMIN-DATA':
                    system.adminData=parseAdminDataNode(xmlElem)
                elif xmlElem.tag=='FIBEX-ELEMENT-REFS':
                    self.parseFibexElementRefs(xmlElem,system)
                elif xmlElem.tag=='MAPPING':
                    self.parseSystemMapping(xmlElem,system)
                elif xmlElem.tag=='SOFTWARE-COMPOSITION':
                    self.parseSoftwareComposition(xmlElem,system)
                else:
                    raise NotImplementedError(xmlElem.tag)
            return system
        else:
            raise KeyError('expected to find <SHORT-NAME> inside <SYSTEM> tag')

    def parseFibexElementRefs(self,xmlRoot,system):
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='FIBEX-ELEMENT-REF':
                system.fibexElementRefs.append(parseTextNode(xmlElem))
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseSystemMapping(self,xmlRoot,system):
        """parses <MAPPING>"""
        assert(xmlRoot.tag=='MAPPING')
        name=parseTextNode(xmlRoot.find('SHORT-NAME'))
        system.mapping = Mapping(name,system)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SHORT-NAME':
                pass
            elif xmlElem.tag=='DATA-MAPPINGS':
                self.parseDataMapping(xmlElem,system.mapping.data)
            elif xmlElem.tag=='SW-IMPL-MAPPINGS':
                self.parseSwImplMapping(xmlElem,system.mapping)
            elif xmlElem.tag=='SW-MAPPINGS':
                self.parseSwMapping(xmlElem,system.mapping)
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseDataMapping(self,xmlRoot,dataMapping):
        """parses <DATA-MAPPINGS>"""
        assert(xmlRoot.tag=='DATA-MAPPINGS')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SWC-TO-IMPL-MAPPING':
                pass
            elif xmlElem.tag=='SENDER-RECEIVER-TO-SIGNAL-MAPPING':
                dataMapping.senderReceiverToSignal.append(self.parseSenderReceiverToSignalMapping(xmlElem))
            elif xmlElem.tag=='SENDER-RECEIVER-TO-SIGNAL-GROUP-MAPPING':
                dataMapping.senderReceiverToSignalGroup.append(self.parseSenderReceiverToSignalGroupMapping(xmlElem))
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseSwImplMapping(self,xmlRoot,mapping):
        """parses <SW-IMPL-MAPPINGS>"""
        assert(xmlRoot.tag=='SW-IMPL-MAPPINGS')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SWC-TO-IMPL-MAPPING':
                pass
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseSwMapping(self,xmlRoot,mapping):
        """parses <SW-MAPPINGS>"""
        assert(xmlRoot.tag=='SW-MAPPINGS')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SWC-TO-ECU-MAPPING':
                pass
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseSoftwareComposition(self,xmlRoot,system):
        """parses <SOFTWARE-COMPOSITION>"""
        assert(xmlRoot.tag=='SOFTWARE-COMPOSITION')
        name=parseTextNode(xmlRoot.find('SHORT-NAME'))
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SHORT-NAME':
                pass
            elif xmlElem.tag=='SOFTWARE-COMPOSITION-TREF':
                pass
            else:
                raise NotImplementedError(xmlElem.tag)

    def parseSenderReceiverToSignalMapping(self,xmlRoot):
        """parses <'SENDER-RECEIVER-TO-SIGNAL-MAPPING'>"""
        assert(xmlRoot.tag=='SENDER-RECEIVER-TO-SIGNAL-MAPPING')
        dataElemIRef=None
        signalRef=None
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='DATA-ELEMENT-IREF':
                dataElemIRef=self.parseDataElemInstanceRef(xmlElem)
            elif xmlElem.tag=='SIGNAL-REF':
                signalRef=parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (dataElemIRef is not None) and (signalRef is not None):
            return SenderReceiverToSignalMapping(dataElemIRef,signalRef)
        else:
            raise Exception("failed to parse SENDER-RECEIVER-TO-SIGNAL-MAPPING")

    def parseSenderReceiverToSignalGroupMapping(self,xmlRoot):
        """parses <'SENDER-RECEIVER-TO-SIGNAL-GROUP-MAPPING'>"""
        assert(xmlRoot.tag=='SENDER-RECEIVER-TO-SIGNAL-GROUP-MAPPING')
        dataElemIRef=None
        signalGroupRef=None
        typeMapping=None
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'DATA-ELEMENT-IREF': #minOccurs=0, maxOccurs=1
                dataElemIRef=self.parseDataElemInstanceRef(xmlElem)
            elif xmlElem.tag == 'SIGNAL-GROUP-REF': #minOccurs=0, maxOccurs=
                signalGroupRef=parseTextNode(xmlElem)
            elif xmlElem.tag == 'TYPE-MAPPING': #minOccurs=0, maxOccurs=1
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag=='SENDER-REC-ARRAY-TYPE-MAPPING':
                        typeMapping = ArrayElementMapping()
                        for xmlItem in xmlChild.findall('./ARRAY-ELEMENT-MAPPINGS/SENDER-REC-ARRAY-TYPE-MAPPING'):
                            print("SENDER-REC-ARRAY-TYPE-MAPPING not implemented")
                    elif xmlChild.tag=='SENDER-REC-RECORD-TYPE-MAPPING':
                        typeMapping=RecordElementMapping()
                        for xmlItem in xmlChild.findall('./RECORD-ELEMENT-MAPPINGS/SENDER-REC-RECORD-ELEMENT-MAPPING'):
                            typeMapping.elements.append(self.parseSenderRecRecordElementMapping(xmlItem))
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        return SenderReceiverToSignalGroupMapping(dataElemIRef,signalGroupRef,typeMapping)

    def parseDataElemInstanceRef(self,xmlRoot):
        dataElemRef=parseTextNode(xmlRoot.find('DATA-ELEMENT-REF'))
        assert(dataElemRef is not None)
        dataElemIRef=SignalDataElementInstanceRef(dataElemRef)
        for xmlChild in xmlRoot.findall('./*'):
            if xmlChild.tag=='DATA-ELEMENT-REF':
                pass
            elif xmlChild.tag=='SOFTWARE-COMPOSITION-REF':
                dataElemIRef.softwareCompositionRef=parseTextNode(xmlChild)
            elif xmlChild.tag=='COMPONENT-PROTOTYPE-REF':
                dataElemIRef.componentPrototypeRef.append(parseTextNode(xmlChild))
            elif xmlChild.tag=='PORT-PROTOTYPE-REF':
                dataElemIRef.portPrototypeRef=parseTextNode(xmlChild)
            else:
                raise NotImplementedError(xmlChild.tag)
        return dataElemIRef

    def parseSenderRecRecordElementMapping(self,xmlRoot):
        """parses <'SENDER-REC-RECORD-ELEMENT-MAPPING'>"""
        assert(xmlRoot.tag=='SENDER-REC-RECORD-ELEMENT-MAPPING')
        recordElementRef=None
        signalRef=None
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='RECORD-ELEMENT-REF': #minOccurs="0" maxOccurs="1"
                recordElementRef=parseTextNode(xmlElem)
            elif xmlElem.tag=='SIGNAL-REF': #minOccurs="0" maxOccurs="1"
                signalRef=parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return SenderRecRecordElementMapping(recordElementRef,signalRef)

    def parseSenderRecArrayElementMapping(self,xmlRoot):
        """parses <'SENDER-REC-RECORD-ELEMENT-MAPPING'>"""
        assert(xmlRoot.tag=='SENDER-REC-RECORD-ELEMENT-MAPPING')
        raise NotImplementedError(xmlRoot.tag)



from autosar.writer.writer_base import ElementWriter
import autosar.constant

class XMLConstantWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

    def getSupportedXML(self):
        return ['Constant']

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        if type(elem).__name__ == 'Constant':
            return self.writeConstantXML(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeConstantXML(self,elem):
        lines = []
        assert(isinstance(elem,autosar.constant.Constant))
        lines.append('<CONSTANT-SPECIFICATION>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        if self.version>=4.0:
            lines.extend(self.indent(self._writeValueXMLV4(elem.value),1))
        else:
            lines.extend(self.indent(self._writeValueXMLV3(elem.value),1))
        lines.append('</CONSTANT-SPECIFICATION>')
        return lines

    def _writeValueXMLV3(self,elem):
        lines=[]
        lines.append('<VALUE>')
        lines.extend(self.indent(self._writeLiteralValueXML(elem),1))
        lines.append('</VALUE>')
        return lines

    def _writeLiteralValueXML(self,elem):
        if isinstance(elem,autosar.constant.IntegerValue):
            return self._writeIntegerLiteralXML(elem)
        elif isinstance(elem,autosar.constant.RecordValue):
            return self._writeRecordSpecificationXML(elem)
        elif isinstance(elem,autosar.constant.StringValue):
            return self._writeStringLiteralXML(elem)
        elif isinstance(elem,autosar.constant.BooleanValue):
            return self._writeBooleanLiteralXML(elem)
        elif isinstance(elem,autosar.constant.ArrayValue):
            return self._writeArraySpecificationXML(elem)
        else:
            raise NotImplementedError(type(elem))

    def _writeIntegerLiteralXML(self,elem):
        assert(isinstance(elem,autosar.constant.IntegerValue))
        lines=[]
        lines.append('<INTEGER-LITERAL>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tag = elem.rootWS().find(elem.typeRef).tag(self.version)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(tag,elem.typeRef),1))
        lines.append(self.indent('<VALUE>%d</VALUE>'%elem.value,1))
        lines.append('</INTEGER-LITERAL>')
        return lines

    def _writeRecordSpecificationXML(self,elem):
        assert(isinstance(elem,autosar.constant.RecordValue))
        lines=[]
        lines.append('<RECORD-SPECIFICATION>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tag = elem.rootWS().find(elem.typeRef).tag(self.version)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(tag,elem.typeRef),1))
        if len(elem.elements)==0: lines.append('<ELEMENTS/>')
        else:
            lines.append(self.indent('<ELEMENTS>',1))
            for childElem in elem.elements:
                lines.extend(self.indent(self._writeLiteralValueXML(childElem),2))
            lines.append(self.indent('</ELEMENTS>',1))
        lines.append('</RECORD-SPECIFICATION>')
        return lines

    def _writeStringLiteralXML(self,elem):
        assert(isinstance(elem,autosar.constant.StringValue))
        lines=[]
        lines.append('<STRING-LITERAL>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tag = elem.rootWS().find(elem.typeRef).tag(self.version)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(tag,elem.typeRef),1))
        lines.append(self.indent('<VALUE>%s</VALUE>'%elem.value,1))
        lines.append('</STRING-LITERAL>')
        return lines

    def _writeBooleanLiteralXML(self,elem):
        assert(isinstance(elem,autosar.constant.BooleanValue))
        lines=[]
        lines.append('<BOOLEAN-LITERAL>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tag = elem.rootWS().find(elem.typeRef).tag(self.version)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(tag,elem.typeRef),1))
        lines.append(self.indent('<VALUE>%s</VALUE>'%('true' if elem.value is True else 'false'),1))
        lines.append('</BOOLEAN-LITERAL>')
        return lines

    def _writeArraySpecificationXML(self,elem):
        assert(isinstance(elem,autosar.constant.ArrayValue))
        lines=[]
        lines.append('<ARRAY-SPECIFICATION>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        tag = elem.rootWS().find(elem.typeRef).tag(self.version)
        lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(tag,elem.typeRef),1))
        if len(elem.elements)==0: lines.append('<ELEMENTS/>')
        else:
            lines.append(self.indent('<ELEMENTS>',1))
            for childElem in elem.elements:
                lines.extend(self.indent(self._writeLiteralValueXML(childElem),2))
            lines.append(self.indent('</ELEMENTS>',1))

        lines.append('</ARRAY-SPECIFICATION>')
        return lines

    def _writeValueXMLV4(self, value):
        lines=[]
        lines.append('<VALUE-SPEC>')
        lines.extend(self.indent(self.writeValueSpecificationXML(value),1))
        lines.append('</VALUE-SPEC>')
        return lines

class CodeConstantWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

    def getSupportedXML(self):
        return []

    def getSupportedCode(self):
        return ['Constant']

    def writeElementXML(self, elem):
        raise NotImplementedError('writeElementXML')

    def writeElementCode(self, elem, localvars):
        if type(elem).__name__ == 'Constant':
            return self.writeConstantCode(elem, localvars)
        else:
            return None

    def writeConstantCode(self, constant, localvars):
        lines=[]
        ws=localvars['ws']
        if not isinstance(constant, autosar.constant.Constant):
            raise ValueError('expected type autosar.constant.Constant')
        if constant.value is not None:
            dataType = ws.find(constant.value.typeRef, role='DataType')
            constructor=None
            if dataType is None:
                raise ValueError('invalid reference: '+constant.value.typeRef)
            if isinstance(constant.value, autosar.constant.ArrayValue):
                initValue = self._writeArrayValueConstantCode(constant.value, localvars)
            elif isinstance(constant.value, autosar.constant.IntegerValue):
                initValue = self._writeIntegerValueConstantCode(constant.value, localvars)
            elif isinstance(constant.value, autosar.constant.StringValue):
                initValue = self._writeStringValueConstantCode(constant.value, localvars)
            elif isinstance(constant.value, autosar.constant.BooleanValue):
                initValue = self._writeBooleanValueConstantCode(constant.value, localvars)
            elif isinstance(constant.value, autosar.constant.RecordValue):
                initValue = self._writeRecordValueConstantCode(constant.value, localvars)
            else:
                raise ValueError('unknown value type: '+type(constant.value))
            params=[repr(constant.name)]
            if ws.roles['DataType'] is not None:
                params.append(repr(dataType.name)) #use name only
            else:
                params.append(repr(dataType.ref)) #use full reference
            if initValue is not None:
                if isinstance(initValue, list):
                    lines.extend(self.writeDictCode('initValue', initValue))
                    params.append('initValue')
                else:
                    params.append(initValue)
            else:
                print(constant.name)
            if constant.adminData is not None:
                param = self.writeAdminDataCode(constant.adminData, localvars)
                assert(len(param)>0)
                params.append('adminData='+param)
            lines.append("package.createConstant(%s)"%(', '.join(params)))
        return lines

    def _writeArrayValueConstantCode(self, value, localvars):
        ws=localvars['ws']
        assert(isinstance(value, autosar.constant.ArrayValue))
        params=[]
        for elem in value.elements:
            if isinstance(elem, autosar.constant.ArrayValue):
                initValue = self._writeArrayValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.IntegerValue):
                initValue = self._writeIntegerValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.StringValue):
                initValue = self._writeStringValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.BooleanValue):
                initValue = self._writeBooleanValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.RecordValue):
                initValue = self._writeRecordValueConstantCode(elem, localvars)
                if isinstance(initValue, list): initValue="{%s}"%(', '.join(initValue)) #join any inner record init values
            else:
                raise ValueError('unknown value type: '+type(elem.value))
            params.append(initValue)
        if len(params)>0:
            return "[%s]"%(', '.join(params))
        return None

    def _writeIntegerValueConstantCode(self, value, localvars):
        return str(value.value)

    def _writeStringValueConstantCode(self, value, localvars):
        return repr(value.value)

    def _writeBooleanValueConstantCode(self, value, localvars):
        return str(value.value)

    def _writeRecordValueConstantCode(self, value, localvars):
        ws=localvars['ws']
        assert(isinstance(value, autosar.constant.RecordValue))
        params=[]
        for elem in value.elements:
            if isinstance(elem, autosar.constant.ArrayValue):
                initValue = self._writeArrayValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.IntegerValue):
                initValue = self._writeIntegerValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.StringValue):
                initValue = self._writeStringValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.BooleanValue):
                initValue = self._writeBooleanValueConstantCode(elem, localvars)
            elif isinstance(elem, autosar.constant.RecordValue):
                initValue = self._writeRecordValueConstantCode(elem, localvars)
                if isinstance(initValue, list): initValue="{%s}"%(', '.join(initValue)) #join any inner record init values
            else:
                raise ValueError('unknown value type: '+type(elem.value))
            params.append('"%s": %s'%(elem.name, initValue))
        if len(params)>0:
            text = "{%s}"%(', '.join(params))
            if len(text)>200: #line will be way too long
                return params
            else:
                return text
        return None



from autosar.element import Element
import autosar.base
import math
import json
import copy
import collections

class RecordTypeElement(Element):
    """
    (AUTOSAR3)
    Implemenetation of <RECORD-ELEMENT> (found inside <RECORD-TYPE>).

    """
    def tag(self, version=None): return 'RECORD-ELEMENT'

    def __init__(self, name, typeRef, parent = None, adminData = None):
        super().__init__(name, parent, adminData)
        self.typeRef=typeRef

    def __eq__(self,other):
        if self is other: return True
        if type(self) == type(other):
            if self.name == other.name:
                lhs = None if self.typeRef is None else self.rootWS().find(self.typeRef)
                rhs = None if other.typeRef is None else other.rootWS().find(other.typeRef)
                if lhs != rhs:
                    print(self.name,self.typeRef)
                return lhs==rhs
        return False

class CompuScaleElement:
    """
    Implementation of <COMPU-SCALE>
    """
    def tag(self, version=None): return 'COMPU-SCALE'

    def __init__(self, lowerLimit, upperLimit, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', label=None, symbol=None, textValue = None, numerator = None, denominator = None, offset = None, mask = None, adminData=None):
        self.lowerLimit = lowerLimit
        self.upperLimit = upperLimit
        self.lowerLimitType = lowerLimitType
        self.upperLimitType = upperLimitType
        self.symbol = symbol
        self.label = label
        self.adminData = adminData
        self.textValue = textValue
        self.offset = offset
        self.numerator = numerator
        self.denominator = denominator
        self.mask = mask

class Unit(Element):
    """
    Implementation of <UNIT>
    """

    def tag(self, version=None): return 'UNIT'

    def __init__(self, name, displayName, factor=None, offset=None, parent=None):
        super().__init__(name, parent)
        self.displayName=displayName
        self.factor = factor       #only supported in AUTOSAR 4 and above
        self.offset = offset       #only supported in AUTOSAR 4 and above
    def __eq__(self,other):
        if self is other: return True
        if type(self) is type(other):
            if (self.name==other.name) and (self.displayName == other.displayName) and (
               self.factor == other.factor) and (self.offset == other.offset):
                return True
        return False

class DataType(Element):
    """
    Base type for DataType (AUTOSAR3)
    """
    def __init__(self, name, parent=None, adminData=None):
        super().__init__(name, parent, adminData)

    @property
    def isComplexType(self):
        return True if isinstance(self, (RecordDataType, ArrayDataType)) else False


class IntegerDataType(DataType):
    """
    IntegerDataType (AUTOSAR3)
    """
    def tag(self,version=None): return 'INTEGER-TYPE'
    def __init__(self, name, minVal=0, maxVal=0, compuMethodRef=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.minVal = int(minVal)
        self.maxVal = int(maxVal)
        self._minValType = 'CLOSED'
        self._maxValType = 'CLOSED'

        if isinstance(compuMethodRef,str):
            self.compuMethodRef=compuMethodRef
        elif hasattr(compuMethodRef,'ref'):
            self.compuMethodRef=compuMethodRef.ref
        else:
            self.compuMethodRef=None
    @property
    def minValType(self):
        return self._minValType
    @minValType.setter
    def minValueType(self, value):
        if (value != "CLOSED") and (value != "OPEN"):
            raise ValueError('value must be either "CLOSED" or "OPEN"')
        self._minValType=value

    @property
    def maxValType(self):
        return self._maxValType
    @maxValType.setter
    def maxValueType(self, value):
        if (value != "CLOSED") and (value != "OPEN"):
            raise ValueError('value must be either "CLOSED" or "OPEN"')
        self._minValType=value


    def __eq__(self,other):
        if self is other: return True
        if type(other) is type(self):
            if (self.name==other.name) and (self.minVal == other.minVal) and (self.maxVal==other.maxVal):
                if (self.compuMethodRef is not None) and (other.compuMethodRef is not None):
                    return self.rootWS().find(self.compuMethodRef) == other.rootWS().find(other.compuMethodRef)
                elif (self.compuMethodRef is None) and (other.compuMethodRef is None):
                    return True

    def __deepcopy__(self,memo):
        obj=type(self)(self.name,self.minVal,self.maxVal,self.compuMethodRef)
        if self.adminData is not None: obj.adminData=copy.deepcopy(self.adminData,memo)
        return obj

class RecordDataType(DataType):
    """
    RecordDataType (AUTOSAR3)
    """
    def tag(self,version=None): return 'RECORD-TYPE'
    def __init__(self, name, elements=None,  parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.elements = []
        if elements is not None:
            for elem in elements:
                if isinstance(elem, RecordTypeElement):
                    self.elements.append(elem)
                    elem.parent=self
                else:
                    raise ValueError('Element must be an instance of RecordTypeElement')
    def __eq__(self, other):
        if self is other: return True
        if (self.name == other.name) and (len(self.elements)==len(other.elements)):
            for i in range(len(self.elements)):
                if self.elements[i] != other.elements[i]: return False
            return True
        return False

    def __deepcopy__(self,memo):
        obj=type(self)(self.name)
        if self.adminData is not None: obj.adminData=copy.deepcopy(self.adminData,memo)
        for elem in self.elements:
            obj.elements.append(RecordTypeElement(elem.name,elem.typeRef,self))
        return obj




class ArrayDataType(DataType):
    """
    ArrayDataType (AUTOSAR 3)
    """
    def tag(self,version=None): return 'ARRAY-TYPE'

    def __init__(self, name, typeRef, length, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.typeRef = typeRef
        self.length = length

class BooleanDataType(DataType):
    """
    BooleanDataType (AUTOSAR 3)
    """
    def tag(self,version=None): return 'BOOLEAN-TYPE'

    def __init__(self,name, parent=None, adminData=None):
        super().__init__(name, parent, adminData)


class StringDataType(DataType):
    """
    StringDataType (AUTOSAR 3)
    """
    def tag(self,version=None): return 'STRING-TYPE'

    def __init__(self,name,length,encoding, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.length=length
        self.encoding=encoding
    def asdict(self):
        data={'type': self.__class__.__name__,'name':self.name,'encoding':self.encoding,'length':self.length}
        return data
    def __eq__(self,other):
        if self is other: return False
        if type(self) == type(other):
            if (self.name==other.name) and (self.length == other.length) and (self.encoding == other.encoding):
                return True
        return False

    def __deepcopy__(self,memo):
        obj=type(self)(self.name,self.length,self.encoding)
        return obj


class RealDataType(DataType):
    """
    RealDataType (AUTOSAR 3)
    """
    def tag(self,version=None): return 'REAL-TYPE'

    def __init__(self, name, minVal, maxVal, minValType='CLOSED', maxValType='CLOSED', hasNaN=False, encoding='SINGLE', parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.minVal=minVal
        self.maxVal=maxVal
        self.minValType = minValType
        self.maxValType = maxValType
        self.hasNaN=hasNaN
        self.encoding=encoding

class Computation:
    """
    Represents one computation (COMPU-INTERNAL-TO-PHYS or COMPU-PHYS-TO-INTERNAL).
    Contains a list of CompuScaleElement objects as well as an optional defaultValue.
    """
    def __init__(self, defaultValue = None):
        self.elements = [] #list of CompuScaleElement
        self.defaultValue = defaultValue

    @property
    def lowerLimit(self):
        """
        Returns lowerLimit of first element
        """
        if len(self.elements) > 0:
            return self.elements[0].lowerLimit
        else:
            raise KeyError('No elements in Computation object')

    @property
    def upperLimit(self):
        """
        Returns upperLimit of last element
        """
        if len(self.elements) > 0:
            return self.elements[-1].upperLimit
        else:
            raise KeyError('No elements in Computation object')

    def createValueTable(self, elements, autoLabel = True):
        """
        Creates a list of CompuScaleElements based on contents of the elements argument

        When elements is a list of strings:
            Creates one CompuScaleElement per list item and automatically calculates lower and upper limits

        When elements is a list of tuples:
            If 2-tuple: First element is both lowerLimit and upperLimit, second element is textValue.
            If 3-tuple: First element is lowerLimit, second element is upperLimit, third element is textValue.

        autoLabel: automatically creates a <SHORT-LABEL> based on the element.textValue (bool). Default=True
        """
        lowerLimitType, upperLimitType = 'CLOSED', 'CLOSED'
        for elem in elements:
            if isinstance(elem, str):
                limit = len(self.elements)
                (lowerLimit, upperLimit, textValue) = (limit, limit, elem)
            elif isinstance(elem, tuple):
                if len(elem) == 2:
                    (limit, textValue) = elem
                    (lowerLimit, upperLimit, textValue) = (limit, limit, textValue)
                elif len(elem) == 3:
                    lowerLimit, upperLimit, textValue = elem
                else:
                    raise ValueError('invalid length: %d'%len(elem))
            else:
                raise ValueError('type not supported:%s'%str(type(elem)))
            label = textValue if autoLabel else None
            self.elements.append(CompuScaleElement(lowerLimit, upperLimit, lowerLimitType, upperLimitType, textValue = textValue, label = label))

    def createRationalScaling(self, offset, numerator, denominator, lowerLimit, upperLimit, lowerLimitType = 'CLOSED', upperLimitType = 'CLOSED', label = None, symbol = None, adminData = None):
        """
        Creates COMPU-SCALE based on rational scaling
        """
        element = CompuScaleElement(lowerLimit, upperLimit, lowerLimitType, upperLimitType, label = label, symbol = symbol, offset = offset, numerator = numerator, denominator = denominator, adminData = adminData)
        self.elements.append(element)
        return element

    def createBitMask(self, elements, autoLabel = True):
        """
        When elements is a list of tuples:

            If 2-tuple: First element is the bitmask (int), second element is the symbol (str)
        """
        lowerLimitType, upperLimitType = 'CLOSED', 'CLOSED'
        for elem in elements:
            if isinstance(elem, tuple):
                if len(elem) == 2:
                    (mask, symbol) = elem
                    (lowerLimit, upperLimit) = (mask, mask)
                else:
                    raise ValueError('invalid length: %d'%len(elem))
            else:
                raise ValueError('type not supported:%s'%str(type(elem)))
            label = symbol if autoLabel else None
            self.elements.append(CompuScaleElement(lowerLimit, upperLimit, lowerLimitType, upperLimitType, symbol = symbol, label = label, mask = mask))

class CompuMethod(Element):
    """
    CompuMethod class
    """
    def tag(self,version=None): return 'COMPU-METHOD'

    def __init__(self, name, useIntToPhys, usePhysToInt, unitRef = None, category = None, parent = None, adminData = None):
        super().__init__(name, parent, adminData, category)
        self.unitRef = unitRef
        self.intToPhys = None
        self.physToInt = None
        if useIntToPhys:
            self.intToPhys = Computation()
        if usePhysToInt:
            self.physToInt = Computation()

class ConstraintBase:
    def __init__(self, lowerLimit, upperLimit, lowerLimitType, upperLimitType):
        if lowerLimit is not None:
            if isinstance(lowerLimit, str) and lowerLimit != '-INF':
                raise ValueError('Unknown lowerLimit: '+lowerLimit)
            self.lowerLimit = lowerLimit
        if upperLimit is not None:
            if isinstance(upperLimit, str) and upperLimit != 'INF':
                raise ValueError('Unknown upperLimit: '+upperLimit)
            self.upperLimit = upperLimit
        if lowerLimitType == 'CLOSED' or lowerLimitType == 'OPEN':
            self.lowerLimitType = lowerLimitType
        else:
            raise ValueError(lowerLimitType)
        if upperLimitType == 'CLOSED' or upperLimitType == 'OPEN':
            self.upperLimitType = upperLimitType
        else:
            raise ValueError(upperLimitType)

    def check_value(self, value):
        if ((self.lowerLimitType=='CLOSED') and (value<self.lowerLimit)) or ((self.lowerLimitType=='OPEN') and (value<=self.lowerLimit)) :
            raise autosar.base.DataConstraintError('Value {} outside lower data constraint ({}) '.format(str(value), str(self.lowerLimit)))
        if ((self.upperLimitType=='CLOSED') and (value>self.upperLimit)) or ((self.upperLimitType=='OPEN') and (value>=self.upperLimit)) :
            raise autosar.base.DataConstraintError('Value {} outside upper data constraint ({}) '.format(str(value), str(self.upperLimit)))

class InternalConstraint(ConstraintBase):
    def __init__(self, lowerLimit=None, upperLimit=None, lowerLimitType='CLOSED', upperLimitType='CLOSED'):
        super().__init__(lowerLimit, upperLimit, lowerLimitType, upperLimitType)

class PhysicalConstraint(ConstraintBase):
    def __init__(self, lowerLimit=None, upperLimit=None, lowerLimitType='CLOSED', upperLimitType='CLOSED'):
        super().__init__(lowerLimit, upperLimit, lowerLimitType, upperLimitType)

class DataConstraint(Element):
    def tag(self,version=None): return 'DATA-CONSTR'

    def __init__(self, name, rules, constraintLevel=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData)
        self.level = constraintLevel
        self.rules = []
        for rule in rules:
            if rule['type'] == 'internalConstraint':
                self.rules.append(InternalConstraint(lowerLimit=rule['lowerLimit'], upperLimit=rule['upperLimit'], lowerLimitType=rule['lowerLimitType'], upperLimitType=rule['upperLimitType']))
            elif rule['type'] == 'physicalConstraint':
                self.rules.append(PhysicalConstraint(lowerLimit=rule['lowerLimit'], upperLimit=rule['upperLimit'], lowerLimitType=rule['lowerLimitType'], upperLimitType=rule['upperLimitType']))
            else:
                raise NotImplementedError
    @property
    def constraintLevel(self):
        if self.level is None or isinstance(self.level, int):
            return self.level
        else:
            raise ValueError('Unknown constraintLevel: '+str(self.level))

    @property
    def lowerLimit(self):
        if len(self.rules) == 1:
            return self.rules[0].lowerLimit
        else:
            raise NotImplementedError('Only a single constraint rule supported')

    @property
    def upperLimit(self):
        if len(self.rules) == 1:
            return self.rules[0].upperLimit
        else:
            raise NotImplementedError('Only a single constraint rule supported')

    @property
    def lowerLimitType(self):
        if len(self.rules) == 1:
            return self.rules[0].lowerLimitType
        else:
            raise NotImplementedError('Only a single constraint rule supported')

    @property
    def upperLimitType(self):
        if len(self.rules) == 1:
            return self.rules[0].upperLimitType
        else:
            raise NotImplementedError('Only a single constraint rule supported')

    def checkValue(self, v):
        if len(self.rules) == 1:
            self.rules[0].check_value(v)
        else:
            raise NotImplementedError('Only a single rule constraint supported')

    def findByType(self, constraintType = 'internalConstraint'):
        """
        Returns the first constraint of the given constraint type (internalConstraint or physicalConstraint)
        """
        for rule in self.rules:
            if (isinstance(rule, InternalConstraint) and constraintType == 'internalConstraint') or (isinstance(rule, PhysicalConstraint) and constraintType == 'physicalConstraint'):
                return rule

class ImplementationDataType(Element):
    def tag(self, version=None): return 'IMPLEMENTATION-DATA-TYPE'
    def __init__(self, name, variantProps = None, dynamicArraySizeProfile = None, typeEmitter = None, category='VALUE', parent = None, adminData = None):
        super().__init__(name, parent, adminData, category)
        self.dynamicArraySizeProfile = dynamicArraySizeProfile
        self.typeEmitter = typeEmitter
        self.variantProps = []
        self.subElements = []
        self.symbolProps = None
        if isinstance(variantProps, (autosar.base.SwDataDefPropsConditional, autosar.base.SwPointerTargetProps)):
            self.variantProps.append(variantProps)
        elif isinstance(variantProps, collections.abc.Iterable):
            for elem in variantProps:
                if isinstance(elem, (autosar.base.SwDataDefPropsConditional, autosar.base.SwPointerTargetProps)):
                    self.variantProps.append(elem)
                else:
                    raise ValueError('Invalid type: ', type(elem))

    def getArrayLength(self):
        """
        Deprecated, use arraySize property instead
        """
        return self.arraySize


    def getTypeReference(self):
        """
        Deprecated, use implementationTypeRef property instead
        """
        return self.implementationTypeRef

    @property
    def arraySize(self):
        if len(self.subElements)>0:
            return self.subElements[0].arraySize
        else:
            return None

    @property
    def implementationTypeRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].implementationTypeRef
        else:
            raise RuntimeError('ImplementationDataType has no variantProps')

    @property
    def baseTypeRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].baseTypeRef
        else:
            raise RuntimeError('Element has no variantProps set')

    @property
    def dataConstraintRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].dataConstraintRef
        else:
            raise RuntimeError('Element has no variantProps set')

    @property
    def compuMethodRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].compuMethodRef
        else:
            raise RuntimeError('Element has no variantProps set')

    def setSymbolProps(self, name, symbol):
        """
        Sets SymbolProps for this data type

        Arguments:
        name: <SHORT-NAME> (str)
        symbol: <SYMBOL> (str)
        """
        self.symbolProps = autosar.base.SymbolProps( str(name), str(symbol))

class SwBaseType(Element):
    def tag(self, version=None): return 'SW-BASE-TYPE'
    def __init__(self, name, size = None, typeEncoding = None, nativeDeclaration = None, category='FIXED_LENGTH', parent = None, adminData = None):
        super().__init__(name, parent, adminData, category)
        self.size = None if size is None else int(size)
        self.nativeDeclaration = nativeDeclaration
        self.typeEncoding = typeEncoding

class ImplementationDataTypeElement(Element):
    def tag(self, version=None): return 'IMPLEMENTATION-DATA-TYPE-ELEMENT'

    def __init__(self, name, category = None, arraySize = None, arraySizeSemantics = None, variantProps = None, parent = None, adminData = None):
        super().__init__(name, parent, adminData, category)
        self.arraySize = arraySize
        self.variantProps = []
        if arraySize is not None:
            if arraySizeSemantics is not None:
                self.arraySizeSemantics = arraySizeSemantics
            else:
                self.arraySizeSemantics = 'FIXED-SIZE'
        else:
            self.arraySizeSemantics = None
        if variantProps is not None:
            if isinstance(variantProps, (autosar.base.SwDataDefPropsConditional, autosar.base.SwPointerTargetProps)):
                self.variantProps.append(variantProps)
            elif isinstance(variantProps, collections.abc.Iterable):
                for elem in variantProps:
                    if isinstance(elem, (autosar.base.SwDataDefPropsConditional, autosar.base.SwPointerTargetProps)):
                        self.variantProps.append(elem)
                    else:
                        raise ValueError('Invalid type: ', type(elem))

class ApplicationDataType(Element):
    """
    Base type for AUTOSAR application data types (AUTOSAR4)

    Arguments:
    name: <SHORT-NAME> (None or str)
    variantProps: <SW-DATA-DEF-PROPS-VARIANTS> (instance (or list) of autosar.base.SwDataDefPropsConditional)
    category: <CATEGORY> (None or str)
    parent: parent object instance (usually the package it will belong to), (object)
    adminData: <ADMIN-DATA> (instance of autosar.base.AdminData or dict)
    """
    def __init__(self, name, variantProps=None, category=None, parent=None, adminData=None):
        super().__init__(name, parent, adminData, category)
        self.variantProps = []
        if variantProps is not None:
            if isinstance(variantProps, autosar.base.SwDataDefPropsConditional):
                self.variantProps.append(variantProps)
            else:
                self.variantProps = list(variantProps)

    @property
    def compuMethodRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].compuMethodRef
        else:
            raise RuntimeError('Element has no variantProps')

    @property
    def dataConstraintRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].dataConstraintRef
        else:
            raise RuntimeError('Element has no variantProps')

    @property
    def unitRef(self):
        if len(self.variantProps)>0:
            return self.variantProps[0].unitRef
        else:
            raise RuntimeError('Element has no variantProps')

class ApplicationPrimitiveDataType(ApplicationDataType):
    """
    Implementation of <APPLICATION-PRIMITIVE-DATA-TYPE> (AUTOSAR 4)

    Arguments:
    (see base class)
    """
    def tag(self, version): return 'APPLICATION-PRIMITIVE-DATA-TYPE'

    def __init__(self, name, variantProps=None, category=None, parent=None, adminData=None):
        super().__init__(name, variantProps, category, parent, adminData)

class ApplicationArrayDataType(ApplicationDataType):
    """
    Implementation of <APPLICATION-ARRAY-DATA-TYPE> (AUTOSAR 4)

    Arguments:
    element: <ELEMENT> (instance of ApplicationArrayElement)
    """
    def tag(self, version): return 'APPLICATION-ARRAY-DATA-TYPE'

    def __init__(self, name, element, variantProps = None, category = 'ARRAY', parent=None, adminData=None):
        super().__init__(name, variantProps, category, parent, adminData)
        if element is None or isinstance(element, ApplicationArrayElement):
            self.element = element
            element.parent = self
        else:
            raise ValueError("element argument must be None or instance of ApplicationArrayElement")

class ApplicationArrayElement(Element):
    """
    An application array element (AUTOSAR 4).
    This is to be used as the element property of ApplicationArrayDataType.

    arguments:
    name: <SHORT-NAME> (None or str)
    typeRef: <TYPE-TREF> (None or str)
    arraySize: <MAX-NUMBER-OF-ELEMENTS> (None or int)
    sizeHandling: <ARRAY-SIZE-HANDLING> (None or str['ALL-INDICES-DIFFERENT-ARRAY-SIZE', 'ALL-INDICES-SAME-ARRAY-SIZE', 'INHERITED-FROM-ARRAY-ELEMENT-TYPE-SIZE', ])
    sizeSemantics: <ARRAY-SIZE-SEMANTICS> (None or str['FIXED-SIZE', 'VARIABLE-SIZE']])
    """
    def tag(self, version=None): return 'ELEMENT'

    def __init__(self, name = None, typeRef = None, arraySize = None, sizeHandling = None, sizeSemantics = 'FIXED-SIZE', category = 'VALUE', parent = None, adminData = None):
        super().__init__(name, parent, adminData, category)
        self.typeRef = None if typeRef is None else str(typeRef)
        self.arraySize = None if arraySize is None else int(arraySize)
        self.sizeHandling = None if sizeHandling is None else str(sizeHandling)
        self.sizeSemantics = None if sizeSemantics is None else str(sizeSemantics)

class ApplicationRecordDataType(ApplicationDataType):
    """
    Implementation of <APPLICATION-RECORD-DATA-TYPE> (AUTOSAR 4)

    Arguments:
    elements: list of ApplicationRecordElement or None
    """
    def tag(self, version): return 'APPLICATION-RECORD-DATA-TYPE'

    def __init__(self, name, elements = None, variantProps = None, category = None, parent = None, adminData = None):
        super().__init__(name, variantProps, category, parent, adminData)
        if elements is None:
            self.elements = []
        else:
            self.elements = list(elements)

    def append(self, element):
        """
        Append element to self.elements list
        """
        if not isinstance(element, ApplicationRecordElement):
            raise ValueError('element must be an instance of ApplicationRecordElement')
        element.parent = self
        self.elements.append(element)

    def createElement(self, name, typeRef, category = 'VALUE', adminData = None):
        """
        Creates a new instance of ApplicationRecordElement and appends it to internal elements list
        """
        element = ApplicationRecordElement(name, typeRef, category, self, adminData)
        self.elements.append(element)

class ApplicationRecordElement(Element):
    """
    Implements <APPLICATION-RECORD-ELEMENT> (AUTOSAR4)
    """

    def tag(self, version): return 'APPLICATION-RECORD-ELEMENT'

    def __init__(self, name, typeRef, category = None, parent=None, adminData=None):
        super().__init__(name, parent, adminData, category)
        if not isinstance(typeRef, str):
            raise autosar.base.InvalidDataTypeRef(typeRef)
        self.typeRef = typeRef

class DataTypeMappingSet(Element):
    def tag(self, version): return 'DATA-TYPE-MAPPING-SET'

    def __init__(self, name, parent = None, adminData = None):
        super().__init__(name, parent, adminData)
        self.applicationTypeMap = {} #applicationDataTypeRef to implementationDataTypeRef dictionary
        self.modeRequestMap = {} #modeDeclarationGroupRef to implementationDataTypeRef dictionary

    def createDataTypeMapping(self, applicationDataTypeRef, implementationDataTypeRef):
        self.applicationTypeMap[applicationDataTypeRef] = implementationDataTypeRef
        return DataTypeMap(applicationDataTypeRef,  implementationDataTypeRef)

    def createModeRequestMapping(self, modeDeclarationGroupRef, implementationDataTypeRef):
        self.modeRequestMap[modeDeclarationGroupRef] = implementationDataTypeRef
        return ModeRequestTypeMap(modeDeclarationGroupRef,  implementationDataTypeRef)

    def add(self, item):
        if isinstance(item, DataTypeMap):
                self.createDataTypeMapping(item.applicationDataTypeRef, item.implementationDataTypeRef)
        elif isinstance(item, ModeRequestTypeMap):
                self.createModeRequestMapping(item.modeDeclarationGroupRef, item.implementationDataTypeRef)
        else:
            raise ValueError("Item is neither an instance of DataTypeMap or ModeRequestTypeMap")

    def getDataTypeMapping(self, applicationDataTypeRef):
        """
        Returns an instance of DataTypeMap or None if not found.
        """
        implementationDataTypeRef = self.applicationTypeMap.get(applicationDataTypeRef, None)
        if implementationDataTypeRef is not None:
            return DataTypeMap(applicationDataTypeRef,  implementationDataTypeRef)
        return None

    def getModeRequestMapping(self, modeDeclarationGroupRef):
        """
        Returns an instance of DataTypeMap or None if not found.
        """
        implementationDataTypeRef = self.modeRequestMap.get(modeDeclarationGroupRef, None)
        if implementationDataTypeRef is not None:
            return ModeRequestTypeMap(modeDeclarationGroupRef,  implementationDataTypeRef)
        return None

    def findMappedDataTypeRef(self, applicationDataTypeRef):
        """
        Returns a reference (str) to the mapped implementation data type or None if not found.
        """
        return self.applicationTypeMap.get(applicationDataTypeRef, None)

    def findMappedDataType(self, applicationDataTypeRef):
        """
        Returns the instance of the mapped implementation data type.
        This requires that both the DataTypeMappingSet and the implementation data type reference are in the same AUTOSAR workspace.
        """
        implementationDataTypeRef = self.applicationTypeMap.get(applicationDataTypeRef, None)
        if implementationDataTypeRef is not None:
            ws = self.rootWS()
            if ws is None:
                raise RuntimeError("Root workspace not found")
            return ws.find(implementationDataTypeRef)
        return None

    def findMappedModeRequestRef(self, modeDeclarationGroupRef):
        """
        Returns a reference (str) to the mapped implementation data type or None if not found.
        """
        return self.modeRequestMap.get(modeDeclarationGroupRef, None)

    def findMappedModeRequest(self, modeDeclarationGroupRef):
        """
        Returns the instance of the mapped implementation data type.
        This requires that both the DataTypeMappingSet and the implementation data type reference are in the same AUTOSAR workspace.
        """
        implementationDataTypeRef = self.modeRequestMap.get(modeDeclarationGroupRef, None)
        if implementationDataTypeRef is not None:
            ws = self.rootWS()
            if ws is None:
                raise RuntimeError("Root workspace not found")
            return ws.find(implementationDataTypeRef)
        return None

class DataTypeMap:
    """
    Mapping from ApplicationDataType to ImplementationDataType.
    """

    def __init__(self, applicationDataTypeRef, implementationDataTypeRef):
        self.applicationDataTypeRef = applicationDataTypeRef
        self.implementationDataTypeRef = implementationDataTypeRef

    def tag(self, version): return 'DATA-TYPE-MAP'

class ModeRequestTypeMap:
    """
    Mapping from ModeGroupDeclaration to ImplementationDataType.
    """
    def __init__(self, modeDeclarationGroupRef, implementationDataTypeRef):
        self.modeDeclarationGroupRef = modeDeclarationGroupRef
        self.implementationDataTypeRef = implementationDataTypeRef

    def tag(self, version): return 'MODE-REQUEST-TYPE-MAP'




import abc
from collections import deque
from autosar.base import (AdminData, SpecialDataGroup, SpecialData, SwDataDefPropsConditional, SwPointerTargetProps, SymbolProps)
import autosar.element

def _parseBoolean(value):
    if value is None:
        return None
    if isinstance(value,str):
        if value == 'true': return True
        elif value =='false': return False
    raise ValueError(value)

class CommonTagsResult:
    def __init__(self):
        self.reset()
    
    def reset(self):
        self.adminData = None
        self.desc = None
        self.descAttr = None
        self.longName = None
        self.longNameAttr = None
        self.name = None
        self.category = None


class BaseParser:
    def __init__(self,version=None):
        self.version = version
        self.common = deque()
    
    def push(self):
        self.common.append(CommonTagsResult())
    
    def pop(self, obj = None):
        if obj is not None:
            self.applyDesc(obj)
            self.applyLongName(obj)
        self.common.pop()

    
        

    def baseHandler(self, xmlElem):
        """
        Alias for defaultHandler
        """
        self.defaultHandler(xmlElem)
        
    def defaultHandler(self, xmlElem):
        """
        A default handler that parses common tags found under most XML elements
        """
        if xmlElem.tag == 'SHORT-NAME':
            self.common[-1].name = self.parseTextNode(xmlElem)
        elif xmlElem.tag == 'ADMIN-DATA':
            self.common[-1].adminData = self.parseAdminDataNode(xmlElem)
        elif xmlElem.tag == 'CATEGORY':
            self.common[-1].category = self.parseTextNode(xmlElem)
        elif xmlElem.tag == 'DESC':
            self.common[-1].desc, self.common[-1].desc_attr = self.parseDescDirect(xmlElem)
        elif xmlElem.tag == 'LONG-NAME':
            self.common[-1].longName, self.common[-1].longName_attr = self.parseLongNameDirect(xmlElem)
        else:
            raise NotImplementedError(xmlElem.tag)
    
    def applyDesc(self, obj):
        if self.common[-1].desc is not None:            
            obj.desc=self.common[-1].desc
            obj.descAttr=self.common[-1].desc_attr

    def applyLongName(self, obj):
        if self.common[-1].longName is not None:            
            obj.longName=self.common[-1].longName
            obj.longNameAttr=self.common[-1].longName_attr
    
    @property
    def name(self):
        return self.common[-1].name

    @property
    def adminData(self):
        return self.common[-1].adminData

    @property
    def category(self):
        return self.common[-1].category

    @property
    def desc(self):
        return self.common[-1].desc, self.common[-1].descAttr

    def parseLongName(self, xmlRoot, elem):
        xmlDesc = xmlRoot.find('LONG-NAME')
        if xmlDesc is not None:
            L2Xml = xmlDesc.find('L-4')
            if L2Xml is not None:
                L2Text=self.parseTextNode(L2Xml)
                L2Attr=L2Xml.attrib['L']
                elem.desc=L2Text
                elem.descAttr=L2Attr

    def parseLongNameDirect(self, xmlLongName):
        assert(xmlLongName.tag == 'LONG-NAME')
        L2Xml = xmlLongName.find('L-4')
        if L2Xml is not None:
            L2Text=self.parseTextNode(L2Xml)
            L2Attr=L2Xml.attrib['L']
            return (L2Text, L2Attr)
        return (None, None)

    def parseDesc(self, xmlRoot, elem):
        xmlDesc = xmlRoot.find('DESC')
        if xmlDesc is not None:
            L2Xml = xmlDesc.find('L-2')
            if L2Xml is not None:
                L2Text=self.parseTextNode(L2Xml)
                L2Attr=L2Xml.attrib['L']
                elem.desc=L2Text
                elem.descAttr=L2Attr

    def parseDescDirect(self, xmlDesc):
        assert(xmlDesc.tag == 'DESC')
        L2Xml = xmlDesc.find('L-2')
        if L2Xml is not None:
            L2Text=self.parseTextNode(L2Xml)
            L2Attr=L2Xml.attrib['L']
            return (L2Text, L2Attr)
        return (None, None)

    def parseTextNode(self, xmlElem):
        return None if xmlElem is None else xmlElem.text

    def parseIntNode(self, xmlElem):
        return None if xmlElem is None else int(xmlElem.text)

    def parseFloatNode(self, xmlElem):
        return None if xmlElem is None else float(xmlElem.text)

    def parseBooleanNode(self, xmlElem):
        return None if xmlElem is None else _parseBoolean(xmlElem.text)

    def parseNumberNode(self, xmlElem):
        textValue = self.parseTextNode(xmlElem)
        retval = None
        try:
            retval = int(textValue)
        except ValueError:
            try:
                retval = float(textValue)
            except ValueError:
                retval = textValue
        return retval

    def hasAdminData(self, xmlRoot):
        return True if xmlRoot.find('ADMIN-DATA') is not None else False

    def parseAdminDataNode(self, xmlRoot):
        if xmlRoot is None: return None
        assert(xmlRoot.tag=='ADMIN-DATA')
        adminData=AdminData()
        xmlSDGS = xmlRoot.find('./SDGS')
        if xmlSDGS is not None:
            for xmlElem in xmlSDGS.findall('./SDG'):
                SDG_GID=xmlElem.attrib['GID']
                specialDataGroup = SpecialDataGroup(SDG_GID)
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'SD':
                        SD_GID = None
                        TEXT=xmlChild.text
                        try:
                            SD_GID=xmlChild.attrib['GID']
                        except KeyError: pass
                        specialDataGroup.SD.append(SpecialData(TEXT, SD_GID))
                    else:
                        raise NotImplementedError(xmlChild.tag)
                adminData.specialDataGroups.append(specialDataGroup)
        return adminData

    def parseSwDataDefProps(self, xmlRoot):
        assert (xmlRoot.tag == 'SW-DATA-DEF-PROPS')
        variants = []
        for itemXML in xmlRoot.findall('./*'):
            if itemXML.tag == 'SW-DATA-DEF-PROPS-VARIANTS':
                for subItemXML in itemXML.findall('./*'):
                    if subItemXML.tag == 'SW-DATA-DEF-PROPS-CONDITIONAL':
                        variant = self.parseSwDataDefPropsConditional(subItemXML)
                        assert(variant is not None)
                        variants.append(variant)
                    else:
                        raise NotImplementedError(subItemXML.tag)
            else:
                raise NotImplementedError(itemXML.tag)
        return variants if len(variants)>0 else None

    def parseSwDataDefPropsConditional(self, xmlRoot):
        assert (xmlRoot.tag == 'SW-DATA-DEF-PROPS-CONDITIONAL')
        (baseTypeRef, implementationTypeRef, swCalibrationAccess, compuMethodRef, dataConstraintRef,
         swPointerTargetPropsXML, swImplPolicy, swAddressMethodRef, unitRef) = (None, None, None, None, None, None, None, None, None)
        for xmlItem in xmlRoot.findall('./*'):
            if xmlItem.tag == 'BASE-TYPE-REF':
                baseTypeRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'SW-CALIBRATION-ACCESS':
                swCalibrationAccess = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'COMPU-METHOD-REF':
                compuMethodRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'DATA-CONSTR-REF':
                dataConstraintRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'SW-POINTER-TARGET-PROPS':
                swPointerTargetPropsXML = xmlItem
            elif xmlItem.tag == 'IMPLEMENTATION-DATA-TYPE-REF':
                implementationTypeRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'SW-IMPL-POLICY':
                swImplPolicy = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'SW-ADDR-METHOD-REF':
                swAddressMethodRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'UNIT-REF':
                unitRef = self.parseTextNode(xmlItem)
            elif xmlItem.tag == 'ADDITIONAL-NATIVE-TYPE-QUALIFIER':
                pass #implement later
            elif xmlItem.tag == 'SW-CALPRM-AXIS-SET':
                print("[BaseParser] unhandled: %s"%xmlItem.tag)
                pass #implement later
                print("[BaseParser] unhandled: %s"%xmlItem.tag)
            elif xmlItem.tag == 'SW-RECORD-LAYOUT-REF':
                print("[BaseParser] unhandled: %s"%xmlItem.tag)
                pass #implement later
            elif xmlItem.tag == 'INVALID-VALUE':
                print("[BaseParser] unhandled: %s"%xmlItem.tag)
                pass #implement later
            else:
                raise NotImplementedError(xmlItem.tag)
        variant = SwDataDefPropsConditional(baseTypeRef, implementationTypeRef, swAddressMethodRef, swCalibrationAccess, swImplPolicy, None, compuMethodRef, dataConstraintRef, unitRef)
        if swPointerTargetPropsXML is not None:
            variant.swPointerTargetProps = self.parseSwPointerTargetProps(swPointerTargetPropsXML, variant)
        return variant

    def parseSwPointerTargetProps(self, rootXML, parent = None):
        assert (rootXML.tag == 'SW-POINTER-TARGET-PROPS')
        props = SwPointerTargetProps()
        for itemXML in rootXML.findall('./*'):
            if itemXML.tag == 'TARGET-CATEGORY':
                props.targetCategory = self.parseTextNode(itemXML)
            if itemXML.tag == 'SW-DATA-DEF-PROPS':
                props.variants = self.parseSwDataDefProps(itemXML)
        return props

    def parseVariableDataPrototype(self, xmlRoot, parent = None):
        assert(xmlRoot.tag == 'VARIABLE-DATA-PROTOTYPE')
        (typeRef, props_variants, isQueued) = (None, None, False)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                props_variants = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'INIT-VALUE':
                pass #Implement later
            else:
                self.defaultHandler(xmlElem)
        if (self.name is not None) and (typeRef is not None):
            dataElement = autosar.element.DataElement(self.name, typeRef, isQueued, category=self.category, parent = parent, adminData = self.adminData)
            if (props_variants is not None) and len(props_variants) > 0:
                dataElement.setProps(props_variants[0])
            self.pop(dataElement)
            return dataElement
        else:
            self.pop()
            raise RuntimeError('SHORT-NAME and TYPE-TREF must not be None')
        
    
    def parseSymbolProps(self, xmlRoot):
        assert(xmlRoot.tag == 'SYMBOL-PROPS')
        name, symbol = None, None
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SYMBOL':
                symbol = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return SymbolProps(name, symbol)
            
class ElementParser(BaseParser, metaclass=abc.ABCMeta):

    def __init__(self, version=None):
        super().__init__(version)

    @abc.abstractmethod
    def getSupportedTags(self):
        """
        Returns a list of tag-names (strings) that this parser supports.
        A generator returning strings is also OK.
        """
    @abc.abstractmethod
    def parseElement(self, xmlElement, parent = None):
        """
        Invokes the parser

        xmlElem: Element to parse (instance of xml.etree.ElementTree.Element)
        parent: the parent object (usually a package object)
        Should return an object derived from autosar.element.Element
        """



import sys
from autosar.parser.parser_base import ElementParser
import autosar.datatype

class DataTypeParser(ElementParser):
    def __init__(self,version=3.0):
        super().__init__(version)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {'ARRAY-TYPE': self.parseArrayType,
                             'BOOLEAN-TYPE': self.parseBooleanType,
                             'INTEGER-TYPE': self.parseIntegerType,
                             'REAL-TYPE': self.parseRealType,
                             'RECORD-TYPE': self.parseRecordType,
                             'STRING-TYPE': self.parseStringType}
        elif self.version >= 4.0:
            self.switcher = {
               'DATA-CONSTR': self.parseDataConstraint,
               'IMPLEMENTATION-DATA-TYPE': self.parseImplementationDataType,
               'SW-BASE-TYPE': self.parseSwBaseType,
               'DATA-TYPE-MAPPING-SET': self.parseDataTypeMappingSet,
               'APPLICATION-PRIMITIVE-DATA-TYPE': self.parseApplicationPrimitiveDataType,
               'APPLICATION-ARRAY-DATA-TYPE' : self.parseApplicationArrayDataType,
               'APPLICATION-RECORD-DATA-TYPE': self.parseApplicationRecordDataTypeXML,
            }

    def getSupportedTags(self):
        return self.switcher.keys()

    def parseElement(self, xmlElement, parent = None):
        parseFunc = self.switcher.get(xmlElement.tag)
        if parseFunc is not None:
            return parseFunc(xmlElement,parent)
        else:
            return None


    def parseIntegerType(self,root,parent=None):
        if self.version>=3.0:
            name=root.find("./SHORT-NAME").text
            minval = int(root.find("./LOWER-LIMIT").text)
            maxval = int(root.find("./UPPER-LIMIT").text)
            dataDefXML = root.find('./SW-DATA-DEF-PROPS')
            dataType = autosar.datatype.IntegerDataType(name,minval,maxval)
            self.parseDesc(root,dataType)
            if dataDefXML is not None:
                for elem in dataDefXML.findall('./*'):
                    if elem.tag=='COMPU-METHOD-REF':
                        dataType.compuMethodRef=self.parseTextNode(elem)
                    else:
                        raise NotImplementedError(elem.tag)
            return dataType

    def parseRecordType(self,root,parent=None):
        if self.version>=3.0:
            elements = []
            name=root.find("./SHORT-NAME").text
            for elem in root.findall('./ELEMENTS/RECORD-ELEMENT'):
                elemName = self.parseTextNode(elem.find("./SHORT-NAME"))
                elemTypeRef = self.parseTextNode(elem.find("./TYPE-TREF"))
                elements.append(autosar.datatype.RecordTypeElement(elemName, elemTypeRef))
            dataType=autosar.datatype.RecordDataType(name,elements);
            self.parseDesc(root,dataType)
            return dataType

    def parseArrayType(self,root,parent=None):
        if self.version>=3.0:
            name=root.find("./SHORT-NAME").text
            length=int(root.find('ELEMENT/MAX-NUMBER-OF-ELEMENTS').text)
            typeRef=root.find('ELEMENT/TYPE-TREF').text
            dataType=autosar.datatype.ArrayDataType(name,typeRef,length)
            self.parseDesc(root,dataType)
            return dataType;

    def parseBooleanType(self,root,parent=None):
        if self.version>=3:
            name=root.find("./SHORT-NAME").text
            dataType=autosar.datatype.BooleanDataType(name)
            self.parseDesc(root,dataType)
            return dataType

    def parseStringType(self,root,parent=None):
        if self.version>=3.0:
            name=root.find("./SHORT-NAME").text

            length=int(root.find('MAX-NUMBER-OF-CHARS').text)
            encoding=root.find('ENCODING').text
            dataType=autosar.datatype.StringDataType(name,length,encoding)
            self.parseDesc(root,dataType)
            return dataType

    def parseRealType(self,root,parent=None):
        if self.version>=3.0:
            name=root.find("./SHORT-NAME").text

            elem = root.find("./LOWER-LIMIT")
            if elem is not None:
                minval = elem.text
                minvalType = elem.attrib['INTERVAL-TYPE']
            elem = root.find("./UPPER-LIMIT")
            if elem is not None:
                maxval = elem.text
                maxvalType = elem.attrib['INTERVAL-TYPE']
            hasNaNText = self.parseTextNode(root.find("./ALLOW-NAN"))
            hasNaN = True if (hasNaNText is not None and hasNaNText == 'true') else False
            encoding = self.parseTextNode(root.find("./ENCODING"))
            dataType=autosar.datatype.RealDataType(name,minval,maxval,minvalType,maxvalType,hasNaN,encoding)
            self.parseDesc(root,dataType)
            return dataType

    def parseDataConstraint(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'DATA-CONSTR')
        rules=[]
        constraintLevel = None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'DATA-CONSTR-RULES':
                for xmlChildElem in xmlElem.findall('./DATA-CONSTR-RULE/*'):
                    if xmlChildElem.tag == 'INTERNAL-CONSTRS':
                        rules.append(self._parseDataConstraintRule(xmlChildElem, 'internalConstraint'))
                    elif xmlChildElem.tag == 'PHYS-CONSTRS':
                        rules.append(self._parseDataConstraintRule(xmlChildElem, 'physicalConstraint'))
                    elif xmlChildElem.tag == 'CONSTR-LEVEL':
                        constraintLevel = self.parseIntNode(xmlChildElem)
                    else:
                        raise NotImplementedError(xmlChildElem.tag)
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.DataConstraint(self.name, rules, constraintLevel, parent, self.adminData)
        self.pop(elem)
        return elem

    def _parseDataConstraintRule(self, xmlElem, constraintType):
        lowerLimitXML = xmlElem.find('./LOWER-LIMIT')
        upperLimitXML = xmlElem.find('./UPPER-LIMIT')
        lowerLimit = None if lowerLimitXML is None else self.parseNumberNode(lowerLimitXML)
        upperLimit = None if upperLimitXML is None else self.parseNumberNode(upperLimitXML)
        lowerLimitType = 'CLOSED'
        upperLimitType = 'CLOSED'
        key = 'INTERVAL-TYPE'
        if lowerLimitXML is not None and key in lowerLimitXML.attrib and lowerLimitXML.attrib[key]=='OPEN':
            lowerLimitType='OPEN'
        if upperLimitXML is not None and key in upperLimitXML.attrib and upperLimitXML.attrib[key]=='OPEN':
            upperLimitType='OPEN'
        return {
            'type': constraintType,
            'lowerLimit': lowerLimit,
            'upperLimit': upperLimit,
            'lowerLimitType': lowerLimitType,
            'upperLimitType': upperLimitType}


    def parseImplementationDataType(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'IMPLEMENTATION-DATA-TYPE')
        variantProps, typeEmitter, parseTextNode, dynamicArraySizeProfile, subElementsXML, symbolProps = None, None, None, None, None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variantProps = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'TYPE-EMITTER':
                typeEmitter = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DYNAMIC-ARRAY-SIZE-PROFILE':
                dynamicArraySizeProfile = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SUB-ELEMENTS':
                subElementsXML = xmlElem
            elif xmlElem.tag == 'SYMBOL-PROPS':
                symbolProps = self.parseSymbolProps(xmlElem)
            else:
                self.defaultHandler(xmlElem)

        dataType = autosar.datatype.ImplementationDataType(
            self.name,
            variantProps = variantProps,
            dynamicArraySizeProfile = dynamicArraySizeProfile,
            typeEmitter = typeEmitter,
            category = self.category,
            parent = parent,
            adminData = self.adminData
            )
        if subElementsXML is not None:
            dataType.subElements = self.parseImplementationDataTypeSubElements(subElementsXML, dataType)
        if symbolProps is not None:
            dataType.symbolProps = symbolProps
        self.pop(dataType)
        return dataType

    def parseImplementationDataTypeSubElements(self, xmlRoot, parent):
        assert (xmlRoot.tag == 'SUB-ELEMENTS')
        elements = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'IMPLEMENTATION-DATA-TYPE-ELEMENT':
                elements.append(self.parseImplementationDataTypeElement(xmlElem, parent))
            else:
                raise NotImplementedError(xmlElem.tag)
        return elements

    def parseImplementationDataTypeElement(self, xmlRoot, parent):
        assert (xmlRoot.tag == 'IMPLEMENTATION-DATA-TYPE-ELEMENT')
        (arraySize, arraySizeSemantics, variants) = (None, None, None)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variants = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'ARRAY-SIZE':
                arraySize = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ARRAY-SIZE-SEMANTICS':
                arraySizeSemantics = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SUB-ELEMENTS':
                pass #implement later
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.ImplementationDataTypeElement(self.name, self.category, arraySize, arraySizeSemantics, variants, parent, self.adminData)
        self.pop(elem)
        return elem


    def parseSwBaseType(self, xmlRoot, parent = None):
        assert (xmlRoot.tag == 'SW-BASE-TYPE')
        baseTypeSize, baseTypeEncoding, nativeDeclaration = None, None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'BASE-TYPE-SIZE':
                baseTypeSize = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'BASE-TYPE-ENCODING':
                baseTypeEncoding = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'NATIVE-DECLARATION':
                nativeDeclaration = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'MEM-ALIGNMENT':
                pass #implement later
            elif xmlElem.tag == 'BYTE-ORDER':
                pass #implement later
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.SwBaseType(self.name, baseTypeSize, baseTypeEncoding, nativeDeclaration, self.category, parent, self.adminData)
        self.pop(elem)
        return elem

    def parseDataTypeMappingSet(self, xmlRoot, parent = None):
        assert (xmlRoot.tag == 'DATA-TYPE-MAPPING-SET')
        (name, dataTypeMaps, adminData) = (None, None, None)
        dataTypeMaps = []
        modeRequestTypeMaps = []
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ADMIN-DATA':
                adminData=self.parseAdminDataNode(xmlElem)
            elif xmlElem.tag == 'SHORT-NAME':
                name = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'DATA-TYPE-MAPS':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'DATA-TYPE-MAP':
                        dataTypeMap = self._parseDataTypeMapXML(xmlChild)
                        assert(dataTypeMap is not None)
                        dataTypeMaps.append(dataTypeMap)
                    else:
                        raise NotImplementedError(xmlElem.tag)
            elif xmlElem.tag == 'MODE-REQUEST-TYPE-MAPS':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'MODE-REQUEST-TYPE-MAP':
                        modeRequestTypeMap = self._parseModeRequestTypeMapXML(xmlChild)
                        assert(modeRequestTypeMap is not None)
                        modeRequestTypeMaps.append(modeRequestTypeMap)
                    else:
                        raise NotImplementedError(xmlElem.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is None):
            raise RuntimeError('SHORT-NAME cannot be None')
        elem = autosar.datatype.DataTypeMappingSet(name, parent, adminData)
        for mapping in dataTypeMaps + modeRequestTypeMaps:
            elem.add(mapping)
        return elem

    def parseApplicationPrimitiveDataType(self, xmlRoot, parent = None):
        assert (xmlRoot.tag == 'APPLICATION-PRIMITIVE-DATA-TYPE')
        variantProps = None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variantProps = self.parseSwDataDefProps(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        if (self.name is None):
            raise RuntimeError('SHORT-NAME cannot be None')
        elem = autosar.datatype.ApplicationPrimitiveDataType(self.name, variantProps, self.category, parent, self.adminData)
        self.pop(elem)
        return elem

    def parseApplicationArrayDataType(self, xmlRoot, parent = None):
        assert (xmlRoot.tag == 'APPLICATION-ARRAY-DATA-TYPE')
        element, variantProps = None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ELEMENT':
                element = self.parseApplicationArrayElement(xmlElem)
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variantProps = self.parseSwDataDefProps(xmlElem)
            elif xmlElem.tag == 'DYNAMIC-ARRAY-SIZE-PROFILE':
                pass #implement later
            else:
                self.defaultHandler(xmlElem)
        if element is None:
            raise RuntimeError('No <ELEMENT> object found')
        elem = autosar.datatype.ApplicationArrayDataType(self.name, element, variantProps, self.category, parent, self.adminData)
        self.pop(elem)
        return elem

    def parseApplicationArrayElement(self, xmlRoot):
        assert (xmlRoot.tag == 'ELEMENT')
        (typeRef, arraySize, sizeHandling, sizeSemantics) = (None, None, None, None)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ARRAY-SIZE-HANDLING':
                sizeHandling = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ARRAY-SIZE-SEMANTICS':
                sizeSemantics = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'MAX-NUMBER-OF-ELEMENTS':
                arraySize = self.parseTextNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.ApplicationArrayElement(self.name, typeRef, arraySize, sizeHandling, sizeSemantics, self.category, adminData = self.adminData)
        self.pop(elem)
        return elem

    def parseApplicationRecordDataTypeXML(self, xmlRoot, parent = None):
        assert (xmlRoot.tag == 'APPLICATION-RECORD-DATA-TYPE')
        elementsXML, variantProps = None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ELEMENTS':
                elementsXML = xmlElem
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variantProps = self.parseSwDataDefProps(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.ApplicationRecordDataType(self.name, None, variantProps, self.category, parent, self.adminData)
        if elementsXML is not None:
            for xmlChild in elementsXML.findall('./'):
                if xmlChild.tag == 'APPLICATION-RECORD-ELEMENT':
                    elem.elements.append(self._parseApplicationRecordElementXML(xmlChild, parent = elem))
                else:
                    raise NotImplementedError(xmlChild.tag)
        self.pop(elem)
        return elem

    def _parseApplicationRecordElementXML(self, xmlRoot, parent):
        assert (xmlRoot.tag == 'APPLICATION-RECORD-ELEMENT')
        typeRef, variantProps = None, None
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'TYPE-TREF':
                typeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'SW-DATA-DEF-PROPS':
                variantProps = self.parseSwDataDefProps(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        elem = autosar.datatype.ApplicationRecordElement(self.name, typeRef, self.category, parent, self.adminData)
        self.pop(elem)
        return elem


    def _parseDataTypeMapXML(self, xmlRoot):
        assert (xmlRoot.tag == 'DATA-TYPE-MAP')
        (applicationDataTypeRef, implementationDataTypeRef) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'APPLICATION-DATA-TYPE-REF':
                applicationDataTypeRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'IMPLEMENTATION-DATA-TYPE-REF':
                implementationDataTypeRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.datatype.DataTypeMap(applicationDataTypeRef, implementationDataTypeRef)

    def _parseModeRequestTypeMapXML(self, xmlRoot):
        assert (xmlRoot.tag == 'MODE-REQUEST-TYPE-MAP')
        (modeDeclarationGroupRef, implementationDataTypeRef) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'MODE-GROUP-REF':
                modeDeclarationGroupRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'IMPLEMENTATION-DATA-TYPE-REF':
                implementationDataTypeRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.datatype.ModeRequestTypeMap(modeDeclarationGroupRef, implementationDataTypeRef)

class DataTypeSemanticsParser(ElementParser):
    def __init__(self,version=3.0):
        super().__init__(version)

    def getSupportedTags(self):
        return ['COMPU-METHOD']

    def parseElement(self, xmlElement, parent = None):
        if xmlElement.tag == 'COMPU-METHOD':
            return self._parseCompuMethodXML(xmlElement, parent)
        else:
            return None

    def _parseCompuMethodXML(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'COMPU-METHOD')
        compuInternalToPhys, compuPhysToInternal, unitRef = None, None, None

        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'COMPU-INTERNAL-TO-PHYS':
                compuInternalToPhys = self._parseComputationXML(xmlElem)
                assert(compuInternalToPhys is not None)
            elif xmlElem.tag == 'COMPU-PHYS-TO-INTERNAL':
                compuPhysToInternal = self._parseComputationXML(xmlElem)
                assert(compuPhysToInternal is not None)
            elif xmlElem.tag == 'UNIT-REF':
                unitRef = self.parseTextNode(xmlElem)
            else:
                self.defaultHandler(xmlElem)
        compuMethod = autosar.datatype.CompuMethod(self.name, False, False, unitRef, self.category, parent, self.adminData)
        self.pop(compuMethod)
        if compuInternalToPhys is not None:
            compuMethod.intToPhys = compuInternalToPhys
        if compuPhysToInternal is not None:
            compuMethod.physToInt = compuPhysToInternal
        return compuMethod

    def _parseComputationXML(self, xmlRoot):
        assert (xmlRoot.tag == 'COMPU-INTERNAL-TO-PHYS') or (xmlRoot.tag == 'COMPU-PHYS-TO-INTERNAL')
        computation = autosar.datatype.Computation()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'COMPU-SCALES':
                for compuScaleXml in xmlElem.findall('COMPU-SCALE'):
                    compuScale = self._parseCompuScaleXML(compuScaleXml)
                    computation.elements.append(compuScale)
            elif xmlElem.tag == 'COMPU-DEFAULT-VALUE':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'V':
                        computation.defaultValue = self.parseNumberNode(xmlChild)
                        break
                    elif xmlChild.tag == 'VT':
                        computation.defaultValue = self.parseTextNode(xmlChild)
                        break
                    elif xmlChild.tag == 'VF':
                        computation.defaultValue = self.parseNumberNode(xmlChild)
                        break
                    else:
                        raise NotImplementedError(xmlChild.tag)
            else:
                raise NotImplementedError(xmlElem.tag)
        return computation

    def _parseCompuScaleXML(self, xmlRoot):
        assert(xmlRoot.tag == 'COMPU-SCALE')
        label, lowerLimit, upperLimit, lowerLimitType, upperLimitType, symbol, adminData = None, None, None, None, None, None, None
        offset, numerator, denominator, textValue, mask = None, None, None, None, None

        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'DESC':
                pass #implement later
            elif xmlElem.tag == 'SHORT-LABEL':
                label = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'LOWER-LIMIT':
                lowerLimit = self.parseNumberNode(xmlElem)
                if (self.version >= 4.0) and 'INTERVAL-TYPE' in xmlElem.attrib:
                    lowerLimitType = xmlElem.attrib['INTERVAL-TYPE']
            elif xmlElem.tag == 'UPPER-LIMIT':
                upperLimit = self.parseNumberNode(xmlElem)
                if (self.version >= 4.0) and 'INTERVAL-TYPE' in xmlElem.attrib:
                    upperLimitType = xmlElem.attrib['INTERVAL-TYPE']
            elif xmlElem.tag == 'COMPU-RATIONAL-COEFFS':
                offset, numerator, denominator = self._parseCompuRationalXML(xmlElem)
            elif xmlElem.tag == 'SYMBOL':
                symbol = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'ADMIN-DATA':
                adminData = self.parseAdminDataNode(xmlElem)
            elif xmlElem.tag == 'COMPU-CONST':
                textValue = self.parseTextNode(xmlElem.find('./VT'))
            elif xmlElem.tag == 'MASK':
                mask = self.parseIntNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        compuScale = autosar.datatype.CompuScaleElement(lowerLimit, upperLimit, lowerLimitType, upperLimitType, label, symbol, adminData)
        compuScale.offset = offset
        compuScale.numerator = numerator
        compuScale.denominator = denominator
        compuScale.textValue = textValue
        compuScale.mask = mask
        return compuScale

    def _parseCompuRationalXML(self, xmlRoot):
        assert(xmlRoot.tag == 'COMPU-RATIONAL-COEFFS')
        numXml = xmlRoot.findall('./COMPU-NUMERATOR/V')
        denXml = xmlRoot.findall('./COMPU-DENOMINATOR/V')
        assert(numXml is not None)
        assert(len(numXml) == 2)
        assert(denXml is not None)
        if self.parseTextNode(numXml[0]):
            offset = self.parseNumberNode(numXml[0])
        else:
            offset = 0
        if self.parseTextNode(numXml[1]):
            numerator = self.parseNumberNode(numXml[1])
        else:
            numerator = 1
        denominator = self.parseNumberNode(denXml[0])
        return offset, numerator, denominator

class DataTypeUnitsParser(ElementParser):
    def __init__(self,version=3.0):
        super().__init__(version)


    def getSupportedTags(self):
        return ['UNIT']

    def parseElement(self, xmlElement, parent = None):
        if xmlElement.tag == 'UNIT':
            return self._parseUnit(xmlElement, parent)
        else:
            return None

    def _parseUnit(self, xmlRoot, parent=None):
        assert (xmlRoot.tag == 'UNIT')
        name = self.parseTextNode(xmlRoot.find("./SHORT-NAME"))
        displayName = self.parseTextNode(xmlRoot.find("./DISPLAY-NAME"))
        if self.version>=4.0:
            factor = self.parseTextNode(xmlRoot.find("./FACTOR-SI-TO-UNIT"))
            offset = self.parseTextNode(xmlRoot.find("./OFFSET-SI-TO-UNIT"))
        else:
            (factor,offset) = (None, None)
        return autosar.datatype.Unit(name, displayName, factor, offset, parent)
